@@ -100,13 +100,14 @@ int main (int argc, char * argv[])
   GtsCartesianGrid g;
   GtsSurface * surface;
   gdouble iso;
-  gboolean verbose = FALSE, tetra = FALSE;
+  gboolean verbose = FALSE, tetra = FALSE, dual = FALSE;
   GtsIsoCartesianFunc func = sphere;
 
   /* parse options using getopt */
   while (c != EOF) {
 #ifdef HAVE_GETOPT_LONG
     static struct option long_options[] = {
+      {"dual", no_argument, NULL, 'd'},
       {"tetra", no_argument, NULL, 't'},
       {"barth", no_argument, NULL, 'b'},
       {"clebsch", no_argument, NULL, 'c'},
@@ -114,11 +115,14 @@ int main (int argc, char * argv[])
       {"verbose", no_argument, NULL, 'v'}
     };
     int option_index = 0;
-    switch ((c = getopt_long (argc, argv, "hvbct", 
+    switch ((c = getopt_long (argc, argv, "hvbctd",
 			      long_options, &option_index))) {
 #else /* not HAVE_GETOPT_LONG */
-    switch ((c = getopt (argc, argv, "hvbct"))) {
+    switch ((c = getopt (argc, argv, "hvbctd"))) {
 #endif /* not HAVE_GETOPT_LONG */
+    case 'd': /* dual */
+      dual = TRUE;
+      break;
     case 't': /* tetra */
       tetra = TRUE;
       break;
@@ -133,13 +137,14 @@ int main (int argc, char * argv[])
       break;
     case 'h': /* help */
       fprintf (stderr,
-             "Usage: iso [OPTION] NX NY NZ VAL< file.gts\n"
+             "Usage: iso [OPTION] NX NY NZ VAL > file.gts\n"
 	     "Compute the isosurface of various functions (default is a sphere).\n"
 	     "The size of the cartesian mesh is given by NX, NY and NZ and the isosurface value by VAL.\n"
 	     "\n"
 	     "  -b    --barth    use the Barth decic function\n"
 	     "  -c    --clebsch  use the Clebsch function\n"
 	     "  -t    --tetra    use marching tetrahedra (default is marching cubes)\n"
+             "  -d    --dual     use \"dual\" marching tetrahedra\n"
 	     "  -v    --verbose  print statistics about the surface\n"
 	     "  -h    --help     display this help and exit\n"
 	     "\n"
@@ -196,6 +201,8 @@ int main (int argc, char * argv[])
 			     gts_vertex_class ());
   if (tetra)
     gts_isosurface_tetra (surface, g, func, NULL, iso);
+  else if (dual)
+    gts_isosurface_tetra_bcl (surface, g, func, NULL, iso);
   else
     gts_isosurface_cartesian (surface, g, func, NULL, iso);
 
