@@ -269,8 +269,10 @@ int main (int argc, char * argv[])
     g_slist_foreach (si->edges, (GFunc) write_edge, stdout);
     printf ("}\n");
   }
-  else
+  else {
+    GTS_POINT_CLASS (gts_vertex_class ())->binary = TRUE;
     gts_surface_write (s3, stdout);
+  }
 
   /* destroy surfaces */
   gts_object_destroy (GTS_OBJECT (s1));
