@@ -32,27 +32,17 @@
 #define PI 3.14159265359
 #endif
 
-static void point_translate (GtsPoint * p, GtsVector t)
-{
-  p->x += t[0];
-  p->y += t[1];
-  p->z += t[2];
-}
-
 int main (int argc, char * argv[])
 {
   GtsSurface * s;
   GtsFile * fp;
   GtsMatrix * m;
-  GtsVector translate = { 0., 0., 0. };
   int c = 0;
   gboolean verbose = FALSE;
   gboolean revert = FALSE;
   gboolean normalize = FALSE;
 
-  m = gts_matrix_new (1., 0., 0.,
-		      0., 1., 0.,
-		      0., 0., 1.);
+  m = gts_matrix_identity (NULL);
 
   /* parse options using getopt */
   while (c != EOF) {
@@ -86,9 +76,7 @@ int main (int argc, char * argv[])
       gdouble angle, cosa, sina;
       GtsMatrix * rot, * p;
       
-      rot = gts_matrix_new (1., 0., 0.,
-			    0., 1., 0.,
-			    0., 0., 1.);
+      rot = gts_matrix_identity (NULL);
       angle = atof (optarg)*PI/180.;
       cosa = cos (angle);
       sina = sin (angle);
@@ -104,9 +92,7 @@ int main (int argc, char * argv[])
       gdouble angle, cosa, sina;
       GtsMatrix * rot, * p;
       
-      rot = gts_matrix_new (1., 0., 0.,
-			    0., 1., 0.,
-			    0., 0., 1.);
+      rot = gts_matrix_identity (NULL);
       angle = atof (optarg)*PI/180.;
       cosa = cos (angle);
       sina = sin (angle);
@@ -122,9 +108,7 @@ int main (int argc, char * argv[])
       gdouble angle, cosa, sina;
       GtsMatrix * rot, * p;
       
-      rot = gts_matrix_new (1., 0., 0.,
-			    0., 1., 0.,
-			    0., 0., 1.);
+      rot = gts_matrix_identity (NULL);
       angle = atof (optarg)*PI/180.;
       cosa = cos (angle);
       sina = sin (angle);
@@ -140,9 +124,8 @@ int main (int argc, char * argv[])
       GtsMatrix * scale, * p;
       gdouble s = atof (optarg);
 
-      scale = gts_matrix_new (s,  0., 0.,
-			      0., s,  0.,
-			      0., 0., s);
+      scale = gts_matrix_identity (NULL);
+      scale[0][0] = scale[1][1] = scale[2][2] = s;
       p = gts_matrix_product (m, scale);
       gts_matrix_destroy (scale);
       gts_matrix_destroy (m);
@@ -153,9 +136,8 @@ int main (int argc, char * argv[])
       GtsMatrix * scale, * p;
       gdouble s = atof (optarg);
 
-      scale = gts_matrix_new (s,  0., 0.,
-			      0., 1., 0.,
-			      0., 0., 1.);
+      scale = gts_matrix_identity (NULL);
+      scale[0][0] = s;
       p = gts_matrix_product (m, scale);
       gts_matrix_destroy (scale);
       gts_matrix_destroy (m);
@@ -166,9 +148,8 @@ int main (int argc, char * argv[])
       GtsMatrix * scale, * p;
       gdouble s = atof (optarg);
 
-      scale = gts_matrix_new (1., 0., 0.,
-			      0., s,  0.,
-			      0., 0., 1.);
+      scale = gts_matrix_identity (NULL);
+      scale[1][1] = s;
       p = gts_matrix_product (m, scale);
       gts_matrix_destroy (scale);
       gts_matrix_destroy (m);
@@ -179,9 +160,8 @@ int main (int argc, char * argv[])
       GtsMatrix * scale, * p;
       gdouble s = atof (optarg);
 
-      scale = gts_matrix_new (1., 0., 0.,
-			      0., 1., 0.,
-			      0., 0., s);
+      scale = gts_matrix_identity (NULL);
+      scale[2][2] = s;
       p = gts_matrix_product (m, scale);
       gts_matrix_destroy (scale);
       gts_matrix_destroy (m);
@@ -189,13 +169,13 @@ int main (int argc, char * argv[])
       break;
     }
     case 't': /* tx */
-      translate[0] += atof (optarg);
+      m[0][3] += atof (optarg);
       break;
     case 'u': /* ty */
-      translate[1] += atof (optarg);
+      m[1][3] += atof (optarg);
       break;
     case 'w': /* tz */
-      translate[2] += atof (optarg);
+      m[2][3] += atof (optarg);
       break;
     case 'i': /* revert */
       revert = TRUE;
@@ -208,7 +188,7 @@ int main (int argc, char * argv[])
              "Usage: transform [OPTION] < file.gts\n"
 	     "Apply geometric transformations to the input.\n"
 	     "\n"
-	     "  -r ANGLE  --rx=ANGLE      rotate around x-axis\n"
+	     "  -r ANGLE  --rx=ANGLE      rotate around x-axis (angle in degrees)\n"
 	     "  -m ANGLE  --ry=ANGLE      rotate around y-axis\n"
 	     "  -n ANGLE  --rz=ANGLE      rotate around z-axis\n"
 	     "  -s FACTOR --scale=FACTOR  scale by FACTOR\n"
@@ -252,7 +232,6 @@ int main (int argc, char * argv[])
   if (revert)
     gts_surface_foreach_face (s, (GtsFunc) gts_triangle_revert, NULL);
   gts_surface_foreach_vertex (s, (GtsFunc) gts_point_transform, m);
-  gts_surface_foreach_vertex (s, (GtsFunc) point_translate, translate);
   if (normalize) {
     GtsBBox * bb = gts_bbox_surface (gts_bbox_class (), s);
     gdouble scale = bb->x2 - bb->x1;
@@ -262,13 +241,13 @@ int main (int argc, char * argv[])
     if (bb->z2 - bb->z1 > scale) scale = bb->z2 - bb->z1;
     if (scale > 0.) scale = 1./scale;
     else scale = 1.;
-    translate[0] = - (bb->x1 + bb->x2)/2.;
-    translate[1] = - (bb->y1 + bb->y2)/2.;
-    translate[2] = - (bb->z1 + bb->z2)/2.;
-    gts_surface_foreach_vertex (s, (GtsFunc) point_translate, translate);
-    sc = gts_matrix_new (scale, 0., 0.,
-			 0., scale, 0.,
-                         0., 0., scale);
+    sc = gts_matrix_identity (NULL);
+    sc[0][3] = - (bb->x1 + bb->x2)/2.;
+    sc[1][3] = - (bb->y1 + bb->y2)/2.;
+    sc[2][3] = - (bb->z1 + bb->z2)/2.;
+    gts_surface_foreach_vertex (s, (GtsFunc) gts_point_transform, sc);
+    sc[0][0] = sc[1][1] = sc[2][2] = scale;    
+    sc[0][3] = sc[1][3] = sc[2][3] = 0.;
     gts_surface_foreach_vertex (s, (GtsFunc) gts_point_transform, sc);
     gts_matrix_destroy (sc);
   }
