@@ -33,6 +33,7 @@ int main (int argc, char * argv[])
 {
   GtsSurface * s;
   gboolean verbose = FALSE;
+  gboolean cm = FALSE;
   int c = 0;
   GtsFile * fp;
 
@@ -40,15 +41,19 @@ int main (int argc, char * argv[])
   while (c != EOF) {
 #ifdef HAVE_GETOPT_LONG
     static struct option long_options[] = {
+      {"cm", no_argument, NULL, 'c'},
       {"help", no_argument, NULL, 'h'},
       {"verbose", no_argument, NULL, 'v'}
     };
     int option_index = 0;
-    switch ((c = getopt_long (argc, argv, "hv", 
+    switch ((c = getopt_long (argc, argv, "hvc", 
 			      long_options, &option_index))) {
 #else /* not HAVE_GETOPT_LONG */
-    switch ((c = getopt (argc, argv, "hv"))) {
+    switch ((c = getopt (argc, argv, "hvc"))) {
 #endif /* not HAVE_GETOPT_LONG */
+    case 'c': /* cm */
+      cm = TRUE;
+      break;
     case 'v': /* verbose */
       verbose = TRUE;
       break;
@@ -60,6 +65,7 @@ int main (int argc, char * argv[])
 	     "manifold. Exit unsuccessfully otherwise.\n"
 	     "\n"
 	     "  -v    --verbose  print statistics about the surface\n"
+	     "  -c    --cm       output center of mass and volume\n"
 	     "  -h    --help     display this help and exit\n"
 	     "\n"
 	     "Reports bugs to %s\n",
@@ -98,5 +104,13 @@ int main (int argc, char * argv[])
   /* print volume */
   printf ("%g\n", gts_surface_volume (s));
 
+  if (cm) {
+    GtsVector cm;
+
+    /* print center of mass */
+    gts_surface_center_of_mass (s, cm);
+    printf ("(%g,%g,%g)\n", cm[0], cm[1], cm[2]);
+  }
+
   return 0; /* success */
 }
