@@ -447,17 +447,17 @@ void gts_bbox_point_distance2 (GtsBBox * bb, GtsPoint * p,
  * @bb: a #GtsBBox.
  * @p: a #GtsPoint.
  *
- * Returns: %TRUE if the ray starting at @p and ending at (@p->x,
- * @p->y, +infty) intersects with @bb, %FALSE otherwise.
+ * Returns: %TRUE if the ray starting at @p and ending at (+infty,
+ * @p->y, @p->z) intersects with @bb, %FALSE otherwise.
  */
 gboolean gts_bbox_is_stabbed (GtsBBox * bb, GtsPoint * p)
 {
   g_return_val_if_fail (bb != NULL, FALSE);
   g_return_val_if_fail (p != NULL, FALSE);
 
-  if (p->x < bb->x1 || p->x > bb->x2 ||
+  if (p->x > bb->x2 ||
       p->y < bb->y1 || p->y > bb->y2 ||
-      p->z > bb->z2)
+      p->z < bb->z1 || p->z > bb->z2)
     return FALSE;
   return TRUE;
 }
@@ -1083,8 +1083,8 @@ static void surface_distance_foreach_triangle (GtsTriangle * t,
 
   area = gts_triangle_area (t);
   *total_area += area;
-  range->mean += area*range_triangle.mean;
-  range->stddev += area*range_triangle.mean*range_triangle.mean;
+  range->sum += area*range_triangle.mean;
+  range->sum2 += area*range_triangle.mean*range_triangle.mean;
 }
 
 /**
@@ -1131,12 +1131,15 @@ void gts_bb_tree_surface_distance (GNode * tree,
 			    data);
 
   if (total_area > 0.) {
-    range->stddev = sqrt ((range->stddev - range->mean*range->mean/total_area)
-			  /total_area);
-    range->mean /= total_area;
+    if (range->sum2 - range->sum*range->sum/total_area >= 0.)
+      range->stddev = sqrt ((range->sum2 - range->sum*range->sum/total_area)
+			    /total_area);
+    else
+      range->stddev = 0.;
+    range->mean = range->sum/total_area;
   }
   else
-    range->min = range->max = 0.;
+    range->min = range->max = range->mean = range->stddev = 0.;
 }
 
 static void surface_distance_foreach_boundary (GtsEdge * e,
@@ -1160,8 +1163,8 @@ static void surface_distance_foreach_boundary (GtsEdge * e,
     
     length = gts_point_distance (GTS_POINT (s->v1), GTS_POINT (s->v2));
     *total_length += length;
-    range->mean += length*range_edge.mean;
-    range->stddev += length*range_edge.mean*range_edge.mean;
+    range->sum += length*range_edge.mean;
+    range->sum2 += length*range_edge.mean*range_edge.mean;
   }
 }
 
@@ -1210,11 +1213,14 @@ void gts_bb_tree_surface_boundary_distance (GNode * tree,
 			    data);
 
   if (total_length > 0.) {
-    range->stddev = sqrt ((range->stddev - 
-			   range->mean*range->mean/total_length)
-			  /total_length);
-    range->mean /= total_length;
+    if (range->sum2 - range->sum*range->sum/total_length >= 0.)
+      range->stddev = sqrt ((range->sum2 - 
+			     range->sum*range->sum/total_length)
+			    /total_length);
+    else
+      range->stddev = 0.;
+    range->mean = range->sum/total_length;
   }
   else
-    range->min = range->max = 0.;
+    range->min = range->max = range->mean = range->stddev = 0.;
 }
