@@ -17,6 +17,11 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#ifdef HAVE_CONFIG_H
+# include <config.h>
+#endif
+
+
 #include <math.h>
 #include "gts.h"
 
@@ -35,7 +40,8 @@ static gint find_closest (GtsTriangle * t, gpointer value, gpointer * data)
     GtsPoint * p = data[4];
 
     if (gts_triangle_orientation (t) > 0.) {
-      gdouble d = gts_point_distance2 (p, GTS_POINT (GTS_SEGMENT (t->e1)->v1));
+      GtsPoint * p1 = GTS_POINT (GTS_SEGMENT (t->e1)->v1);
+      gdouble d = (p->x - p1->x)*(p->x - p1->x) + (p->y - p1->y)*(p->y - p1->y);
       
       if (d < *dmin) {
 	*dmin = d;
@@ -67,11 +73,57 @@ static GtsFace * closest_face (GtsSurface * s, GtsPoint * p)
   data[3] = &n;
   data[4] = p;
   g_tree_traverse (s->faces, (GTraverseFunc) find_closest, G_IN_ORDER, data);
+
   return closest;
 }
 
 #else /* not USE_SURFACE_BTREE */
 
+#  if GLIB_CHECK_VERSION(2,4,0)
+/* finally, with g_hash_table_find we are able to stop iteration over the hash 
+   table in the middle */
+
+typedef struct _SFindClosest SFindClosest; 
+
+struct _SFindClosest {
+  gdouble dmin; 
+  GtsFace *closest;
+  GtsPoint * p;
+  gint stop;
+};
+
+static gboolean find_closest (gpointer key, gpointer value, gpointer user_data)
+{
+  SFindClosest * data = (SFindClosest *) user_data;
+  GtsFace * f = GTS_FACE (value);
+  
+  if (gts_triangle_orientation (GTS_TRIANGLE (f)) > 0.) {
+    gdouble d = gts_point_distance2 (data->p, 
+				     GTS_POINT (GTS_SEGMENT (GTS_TRIANGLE (f)->e1)->v1));
+    if (d < data->dmin) {
+      data->dmin = d;
+      data->closest = f;
+    }
+  }
+  data->stop--;
+  return !(data->stop > 0);
+}
+
+static GtsFace * closest_face (GtsSurface * s, GtsPoint * p)
+{
+  SFindClosest fc;
+
+  fc.dmin = G_MAXDOUBLE;
+  fc.closest = NULL;
+  fc.p = p;
+  fc.stop = exp (log ((gdouble) g_hash_table_size (s->faces))/3.);
+  g_hash_table_find (s->faces, find_closest, &fc);
+  
+  return fc.closest;
+}
+
+#  else /* VERSION < 2.4.0 */
+
 /* Due to an unkown reason g_hash_table_foreach does not allow to stop 
  * the loop, hence the redefinition. I hope they don't change
  * the GHashTable, GHashNode structures ... */
@@ -84,7 +136,6 @@ struct _GHashNode
   GHashNode *next;
 };
 
-#if GLIB_CHECK_VERSION(1,2,8)
 struct _GHashTable
 {
   gint size;
@@ -94,23 +145,9 @@ struct _GHashTable
   GHashFunc hash_func;
   GCompareFunc key_compare_func;
 };
-#else  /* not 1.2.8 */
-#  if GLIB_CHECK_VERSION(1,3,0)
-struct _GHashTable
-{
-  gint size;
-  gint nnodes;
-  GHashNode **nodes;
-  GHashFunc hash_func;
-  GEqualFunc key_equal_func;
-};
-#  else  /* not 1.3.0 */
-#    error "Hack needs glib version 1.2.8 or 1.3.0"
-#  endif /* not 1.3.0 */
-#endif /* not 1.2.8 */
 
 /* select the face closest to @p among n^1/3 randomly picked faces
- *  of @surface */
+ * of @surface */
 static GtsFace * closest_face (GtsSurface * s, GtsPoint * p)
 {
   guint i, n, nt, ns;
@@ -122,16 +159,16 @@ static GtsFace * closest_face (GtsSurface * s, GtsPoint * p)
   nt = g_hash_table_size (hash_table);
   if (!nt)
     return NULL;
-  ns = exp(log((gdouble) nt)/3.);
 
+  ns = exp(log((gdouble) nt)/3.);
   for (i = 0, n = 0; i < hash_table->size && n < ns; i++)
     for (node = hash_table->nodes[i]; node && n < ns; node = node->next) {
       GtsFace * f = node->key;
 
       if (gts_triangle_orientation (GTS_TRIANGLE (f)) > 0.) {
-	gdouble d = gts_point_distance2 (p, 
-		      GTS_POINT (GTS_SEGMENT (GTS_TRIANGLE (f)->e1)->v1));
-
+	GtsPoint * p1 = GTS_POINT (GTS_SEGMENT (GTS_TRIANGLE (f)->e1)->v1);
+	gdouble d = (p->x - p1->x)*(p->x - p1->x) + (p->y - p1->y)*(p->y - p1->y);
+		 
 	if (d < dmin) {
 	  dmin = d;
 	  closest = f;
@@ -141,7 +178,7 @@ static GtsFace * closest_face (GtsSurface * s, GtsPoint * p)
     }
   return closest;
 }
-
+#  endif /* VERSION < 2.4.0 */
 #endif /* not USE_SURFACE_BTREE */
 
 /* returns the face belonging to @surface and neighbor of @f via @e */
@@ -215,9 +252,10 @@ static GtsEdge * triangle_next_edge (GtsTriangle * t,
       *on_summit = TRUE;
     return NULL;
   }
-
-  if (gts_point_orientation (GTS_POINT (v2), GTS_POINT (v3), p) < 0.0 ||
-      gts_point_orientation (GTS_POINT (v1), GTS_POINT (v2), p) < 0.0)
+  
+  if (gts_point_orientation (GTS_POINT (v2), GTS_POINT (v3), p) < 0.0)
+    return e2;
+  if (gts_point_orientation (GTS_POINT (v1), GTS_POINT (v2), p) < 0.0)
     *on_summit = TRUE;
   return NULL;
 }
@@ -262,8 +300,9 @@ static GtsFace * point_locate (GtsPoint * o,
   }
   
   f = neighbor (f, prev, surface);
-  gts_triangle_vertices_edges (GTS_TRIANGLE (f), prev, 
-			       &v1, &v2, &v3, &prev, &e2, &e3);
+  if (f)
+    gts_triangle_vertices_edges (GTS_TRIANGLE (f), prev, 
+				 &v1, &v2, &v3, &prev, &e2, &e3);
   while (f) {
     gdouble orient = gts_point_orientation (o, GTS_POINT (v3), p);
 
@@ -347,8 +386,8 @@ GtsFace * gts_point_locate (GtsPoint * p,
   if (guess == NULL)
     guess = closest_face (surface, p);
   else
-    g_return_val_if_fail (gts_triangle_orientation (GTS_TRIANGLE (guess)) > 0.,
-			  NULL);
+    g_return_val_if_fail (gts_triangle_orientation (GTS_TRIANGLE (guess)) > 0., NULL);
+
   if (guess == NULL)
     return NULL;
 
@@ -394,6 +433,30 @@ GtsConstraintClass * gts_constraint_class (void)
   return klass;
 }
 
+static void split_list (GtsListFace * f, GtsListFace * f1, GtsListFace * f2, 
+			GtsPoint * p1, GtsPoint * p2,
+			GSList ** last1, GSList ** last2)
+{
+  GSList * i = f->points, * l1 = *last1, * l2 = *last2;
+
+  while (i) {
+    GtsPoint * p = i->data;
+    
+    if (gts_point_orientation (p1, p2, p) >= 0.) {
+      if (l1) l1->next = i; else f1->points = i;
+      l1 = i;
+    }
+    else {
+      if (l2) l2->next = i; else f2->points = i;
+      l2 = i;
+    }
+    i = i->next;
+  }
+  f->points = NULL;
+  *last1 = l1;
+  *last2 = l2;
+}
+
 /* cf. figure misc/swap.fig */
 static void swap_if_in_circle (GtsFace * f1,
 			       GtsVertex * v1, 
@@ -445,6 +508,20 @@ static void swap_if_in_circle (GtsFace * f1,
     gts_object_attributes (GTS_OBJECT (f3), GTS_OBJECT (f1));
     f4 = gts_face_new (surface->face_class, en, e3, e4);
     gts_object_attributes (GTS_OBJECT (f4), GTS_OBJECT (f2));
+    
+    if (GTS_IS_LIST_FACE (f3)) {
+      GSList * last3 = NULL, * last4 = NULL;
+
+      if (GTS_IS_LIST_FACE (f1))
+	split_list (GTS_LIST_FACE (f1), GTS_LIST_FACE (f3), GTS_LIST_FACE (f4),
+		    GTS_POINT (v3), GTS_POINT (v4), &last3, &last4);
+      if (GTS_IS_LIST_FACE (f2))
+	split_list (GTS_LIST_FACE (f2), GTS_LIST_FACE (f3), GTS_LIST_FACE (f4),
+		    GTS_POINT (v3), GTS_POINT (v4), &last3, &last4);
+      if (last3) last3->next = NULL;
+      if (last4) last4->next = NULL;
+    }
+
     gts_surface_remove_face (surface, f1);
     gts_surface_remove_face (surface, f2);
     gts_surface_add_face (surface, f3);
@@ -476,7 +553,7 @@ GtsVertex * gts_delaunay_add_vertex_to_face (GtsSurface * surface,
   GtsSegment * s4, * s5, * s6;
   GtsEdge * e4, * e5, * e6;
   GtsVertex * v1, * v2, * v3;
-  GtsFace * f1, * f2, * f3;
+  GtsFace * nf[3];
 
   g_return_val_if_fail (surface != NULL, v);
   g_return_val_if_fail (v != NULL, v);
@@ -513,21 +590,59 @@ GtsVertex * gts_delaunay_add_vertex_to_face (GtsSurface * surface,
     e6 = GTS_EDGE (s6);
 
   /* cf. figure misc/swap.fig */
-  f1 = gts_face_new (surface->face_class, e4, e1, e5);
-  gts_object_attributes (GTS_OBJECT (f1), GTS_OBJECT (f));
-  f2 = gts_face_new (surface->face_class, e5, e2, e6);
-  gts_object_attributes (GTS_OBJECT (f2), GTS_OBJECT (f));
-  f3 = gts_face_new (surface->face_class, e6, e3, e4);
-  gts_object_attributes (GTS_OBJECT (f3), GTS_OBJECT (f));
+  nf[0] = gts_face_new (surface->face_class, e4, e1, e5);
+  gts_object_attributes (GTS_OBJECT (nf[0]), GTS_OBJECT (f));
+  nf[1] = gts_face_new (surface->face_class, e5, e2, e6);
+  gts_object_attributes (GTS_OBJECT (nf[1]), GTS_OBJECT (f));
+  nf[2] = gts_face_new (surface->face_class, e6, e3, e4);
+  gts_object_attributes (GTS_OBJECT (nf[2]), GTS_OBJECT (f));
+
+  if (GTS_IS_LIST_FACE (f) && GTS_IS_LIST_FACE (nf[0])) {
+    GSList * i = GTS_LIST_FACE (f)->points, * last[3] = { NULL, NULL, NULL };
+
+    while (i) {
+      GtsPoint * p = i->data;
+      GSList * next = i->next;
+      guint j;
+      
+      if (p != GTS_POINT (v)) {
+	if (gts_point_orientation (GTS_POINT (v), GTS_POINT (v1), p) >= 0.) {
+	  gdouble o = gts_point_orientation (GTS_POINT (v), GTS_POINT (v2), p);
+
+	  if (o != 0.)
+	    j = o > 0. ? 1 : 0;
+	  else
+	    j = gts_point_orientation (GTS_POINT (v), GTS_POINT (v3), p) 
+	      > 0. ? 0 : 1;
+	}
+	else if (gts_point_orientation (GTS_POINT (v), GTS_POINT (v3), p) > 0.)
+	  j = 2;
+	else
+	  j = 1;
+	if (last[j])
+	  last[j]->next = i; 
+	else 
+	  GTS_LIST_FACE (nf[j])->points = i;
+	last[j] = i;
+      }
+      else
+	g_slist_free_1 (i);
+      i = next;
+    }
+    GTS_LIST_FACE (f)->points = NULL;
+    if (last[0]) last[0]->next = NULL;
+    if (last[1]) last[1]->next = NULL;
+    if (last[2]) last[2]->next = NULL;
+  }
 
   gts_surface_remove_face (surface, f);
-  gts_surface_add_face (surface, f1);
-  gts_surface_add_face (surface, f2);
-  gts_surface_add_face (surface, f3);
+  gts_surface_add_face (surface, nf[0]);
+  gts_surface_add_face (surface, nf[1]);
+  gts_surface_add_face (surface, nf[2]);
 
-  swap_if_in_circle (f1, v1, v2, v, e1, e5, e4, surface);
-  swap_if_in_circle (f2, v2, v3, v, e2, e6, e5, surface);
-  swap_if_in_circle (f3, v3, v1, v, e3, e4, e6, surface);
+  swap_if_in_circle (nf[0], v1, v2, v, e1, e5, e4, surface);
+  swap_if_in_circle (nf[1], v2, v3, v, e2, e6, e5, surface);
+  swap_if_in_circle (nf[2], v3, v1, v, e3, e4, e6, surface);
 
   return NULL;
 }
@@ -839,7 +954,10 @@ remove_intersected_vertex (GtsSegment * s,
 	if (o3 >= 0.) /* @s->v2 is inside (or on the edge) of t */
 	  return NULL;
 
+	gts_allow_floating_faces = TRUE;
 	gts_surface_remove_face (surface, GTS_FACE (t));
+	gts_allow_floating_faces = FALSE;
+
 	*left = g_slist_prepend (*left, e2);
 	*right = g_slist_prepend (*right, e1);
 
@@ -932,6 +1050,8 @@ GSList * gts_delaunay_add_constraint (GtsSurface * surface,
   triangulate_polygon (right, surface);
   triangulate_polygon (left, surface);
 #endif
+  if (ref && !ref->surfaces)
+    gts_object_destroy (GTS_OBJECT (ref));
   return constraints;
 }
 
@@ -1027,3 +1147,39 @@ void gts_delaunay_remove_hull (GtsSurface * surface)
   }
   gts_allow_floating_edges = FALSE;
 }
+
+/* GtsListFace: Object */
+
+static void gts_list_face_destroy (GtsObject * object)
+{
+  g_slist_free (GTS_LIST_FACE (object)->points);
+
+  (* GTS_OBJECT_CLASS (gts_list_face_class ())->parent_class->destroy) 
+    (object);
+}
+
+static void gts_list_face_class_init (GtsFaceClass * klass)
+{
+  GTS_OBJECT_CLASS (klass)->destroy = gts_list_face_destroy;
+}
+
+GtsFaceClass * gts_list_face_class (void)
+{
+  static GtsFaceClass * klass = NULL;
+
+  if (klass == NULL) {
+    GtsObjectClassInfo gts_list_face_info = {
+      "GtsListFace",
+      sizeof (GtsListFace),
+      sizeof (GtsFaceClass),
+      (GtsObjectClassInitFunc) gts_list_face_class_init,
+      (GtsObjectInitFunc) NULL,
+      (GtsArgSetFunc) NULL,
+      (GtsArgGetFunc) NULL
+    };
+    klass = gts_object_class_new (GTS_OBJECT_CLASS (gts_face_class ()),
+				  &gts_list_face_info);
+  }
+
+  return klass;
+}
