@@ -748,8 +748,9 @@ GtsWGEdge * gts_wgedge_new (GtsWGEdgeClass * klass,
 
 static void graph_init (GtsGraph * g)
 {
-  g->node_class = gts_gnode_class ();
-  g->edge_class = gts_gedge_class ();
+  g->graph_class = gts_graph_class ();
+  g->node_class  = gts_gnode_class ();
+  g->edge_class  = gts_gedge_class ();
 }
 
 static void graph_write (GtsObject * object, FILE * fp)
@@ -764,22 +765,38 @@ static void graph_write (GtsObject * object, FILE * fp)
 
 static void graph_read (GtsObject ** object, GtsFile * f)
 {
-  if (f->type != GTS_STRING) {
-    gts_file_error (f, "expecting a string (GtsGraphClass)");
-    return;
-  }
-  gts_file_next_token (f);
+  GtsObjectClass * klass;
 
   if (f->type != GTS_STRING) {
     gts_file_error (f, "expecting a string (GtsGNodeClass)");
     return;
   }
+  klass = gts_object_class_from_name (f->token->str);
+  if (klass == NULL) {
+    gts_file_error (f, "unknown class `%s'", f->token->str);
+    return;
+  }
+  if (!gts_object_class_is_from_class (klass, gts_gnode_class ())) {
+    gts_file_error (f, "class `%s' is not a GtsGNodeClass", f->token->str);
+    return;
+  }
+  GTS_GRAPH (*object)->node_class = GTS_GNODE_CLASS (klass);
   gts_file_next_token (f);
 
   if (f->type != GTS_STRING) {
     gts_file_error (f, "expecting a string (GtsGEdgeClass)");
     return;
   }
+  klass = gts_object_class_from_name (f->token->str);
+  if (klass == NULL) {
+    gts_file_error (f, "unknown class `%s'", f->token->str);
+    return;
+  }
+  if (!gts_object_class_is_from_class (klass, gts_gedge_class ())) {
+    gts_file_error (f, "class `%s' is not a GtsGEdgeClass", f->token->str);
+    return;
+  }
+  GTS_GRAPH (*object)->edge_class = GTS_GEDGE_CLASS (klass);
   gts_file_next_token (f);
 }
 
@@ -1220,7 +1237,7 @@ guint gts_graph_read_jostle (GtsGraph * g, GtsFile * fp)
     gts_file_error (fp, "expecting an integer (number of nodes)");
     return fp->line;
   }
-  nn = atoi (fp->token->str);  
+  nn = atoi (fp->token->str);
   gts_file_next_token (fp);
 
   if (fp->type != GTS_INT) {
@@ -1347,43 +1364,69 @@ void gts_graph_write (GtsGraph * g, FILE * fp)
 
 /**
  * gts_graph_read:
- * @g: a #GtsGraph.
  * @fp: a #GtsFile.
  *
- * Adds to @g the data read from @fp. The format of the file pointed
- * to by @fp is as described in gts_graph_write().
+ * Reads a graph from a file.
  *
- * Returns: 0 if successful or the line number at which the parsing
- * stopped in case of error (in which case the @error field of @fp is set).
+ * Returns: the new #GtsGraph or %NULL if an error occured (in which
+ * case the @error field of @fp is set).
  */
-guint gts_graph_read (GtsGraph * g, GtsFile * fp)
+GtsGraph * gts_graph_read (GtsFile * fp)
 {
+  GtsGraph * g;
   GtsGNode ** nodes;
   guint nn, ne, n;
 
-  g_return_val_if_fail (g != NULL, 1);
-  g_return_val_if_fail (fp != NULL, 1);
+  g_return_val_if_fail (fp != NULL, NULL);
 
   if (fp->type != GTS_INT) {
     gts_file_error (fp, "expecting an integer (number of nodes)");
-    return fp->line;
+    return NULL;
   }
   nn = atoi (fp->token->str);
   gts_file_next_token (fp);
 
   if (fp->type != GTS_INT) {
     gts_file_error (fp, "expecting an integer (number of edges)");
-    return fp->line;
+    return NULL;
   }
   ne = atoi (fp->token->str);
 
   gts_file_next_token (fp);
-  if (fp->type != '\n')
-    if (GTS_OBJECT (g)->klass->read)
-      (* GTS_OBJECT (g)->klass->read) ((GtsObject **) &g, fp);
+  if (fp->type != '\n') {
+    GtsObjectClass * klass;
+
+    gts_graph_class ();
+    gts_gnode_class ();
+    gts_gedge_class ();
+
+    if (fp->type != GTS_STRING) {
+      gts_file_error (fp, "expecting a string (GtsGraphClass)");
+      return NULL;
+    }
+    klass = gts_object_class_from_name (fp->token->str);
+    if (klass == NULL) {
+      gts_file_error (fp, "unknown class `%s'", fp->token->str);
+      return NULL;
+    }
+    if (!gts_object_class_is_from_class (klass, gts_graph_class ())) {
+      gts_file_error (fp, "class `%s' is not a GtsGraphClass", fp->token->str);
+      return NULL;
+    }
+    g = GTS_GRAPH (gts_object_new (klass));
+    g->graph_class = GTS_GRAPH_CLASS (klass);
+    gts_file_next_token (fp);
+    (* klass->read) ((GtsObject **) &g, fp);
+    if (fp->type == GTS_ERROR) {
+      gts_object_destroy (GTS_OBJECT (g));
+      return NULL;
+    }
+  }
+  else
+    g = GTS_GRAPH (gts_object_new (GTS_OBJECT_CLASS (gts_graph_class ())));
   gts_file_first_token_after (fp, '\n');
   if (nn <= 0)
-    return 0;
+    return g;
 
   nodes = g_malloc ((nn + 1)*sizeof (GtsGNode *));
 
@@ -1446,9 +1489,11 @@ guint gts_graph_read (GtsGraph * g, GtsFile * fp)
   }
   g_free (nodes);
 
-  if (fp->type == GTS_ERROR)
-    return fp->line;
-  return 0;  
+  if (fp->type == GTS_ERROR) {
+    gts_object_destroy (GTS_OBJECT (g));
+    return NULL;
+  }
+  return g;
 }
 
 static void write_dot_node (GtsGNode * node, gpointer * data)
