@@ -22,7 +22,9 @@
 
 #include <stdio.h>
 #include <glib.h>
-#include <gtsconfig.h>
+#ifndef NATIVE_WIN32
+# include <gtsconfig.h>
+#endif
 #ifdef GTS_COMPILATION
 # include "config.h"
 #endif /* not GTS_COMPILATION */
@@ -58,7 +60,7 @@ GTS_C_VAR const guint gts_binary_age;
     (gts_major_version == (major) && gts_minor_version == (minor) && \
      gts_micro_version >= (micro)))
 
-#define GTS_COMMENTS  "#%!"
+#define GTS_COMMENTS  "#!"
 #define GTS_MAINTAINER "popinet@users.sourceforge.net"
 
 /* Class declarations for base types */
@@ -89,7 +91,8 @@ typedef void         (*GtsArgSetFunc)          (GtsObject * obj);
 typedef void         (*GtsArgGetFunc)          (GtsObject * obj);
 
 typedef gdouble                  GtsVector[3];
-typedef GtsVector                GtsMatrix;
+typedef gdouble                  GtsVector4[4];
+typedef GtsVector4               GtsMatrix;
 typedef gdouble                  (*GtsKeyFunc)    (gpointer item,
 						   gpointer data);
 typedef enum 
@@ -110,11 +113,74 @@ typedef gint   (*GtsFunc)              (gpointer item,
 
 /* misc.c */
 
-gint gts_get_token               (FILE * fptr, 
-				  GString * buf, 
-				  const char * delimiters,
-				  const char * comments);
-void gts_get_newline             (FILE * fptr);
+typedef struct _GtsFile GtsFile;
+
+typedef enum {
+  GTS_NONE   = 1 << 8,
+  GTS_INT    = 1 << 9,
+  GTS_UINT   = 1 << 10,
+  GTS_FLOAT  = 1 << 11,
+  GTS_DOUBLE = 1 << 12,
+  GTS_STRING = 1 << 13,
+  GTS_FILE   = 1 << 14,
+  GTS_ERROR  = 1 << 15
+} GtsTokenType;
+
+struct _GtsFile {
+  FILE * fp;
+  guint line, pos;
+  GString * token;
+  GtsTokenType type;
+  gchar * error;
+
+  guint curline, curpos;
+  guint scope, scope_max;
+  gint next_token;
+  gchar * delimiters;
+  gchar * comments;
+  gchar * tokens;
+};
+
+typedef struct _GtsFileVariable GtsFileVariable;
+
+struct _GtsFileVariable {
+  GtsTokenType type;
+  gchar name[30];
+  gboolean unique;
+  gpointer data;
+  gboolean set;
+  guint line, pos;
+};
+
+
+GtsFile *      gts_file_new               (FILE * fp);
+void           gts_file_verror            (GtsFile * f,
+					   const gchar * format,
+					   va_list args);
+void           gts_file_error             (GtsFile * f,
+					   const gchar * format,
+					   ...);
+gint           gts_file_getc              (GtsFile * f);
+guint          gts_file_read              (GtsFile * f, 
+					   gpointer ptr, 
+					   guint size, 
+					   guint nmemb);
+gint           gts_file_getc_scope        (GtsFile * f);
+void           gts_file_next_token        (GtsFile * f);
+void           gts_file_first_token_after (GtsFile * f, 
+					   GtsTokenType type);
+void           gts_file_assign_start      (GtsFile * f, 
+					   GtsFileVariable * vars);
+GtsFileVariable * gts_file_assign_next    (GtsFile * f, 
+					   GtsFileVariable * vars);
+void           gts_file_assign_variables  (GtsFile * f, 
+					   GtsFileVariable * vars);
+void           gts_file_variable_error    (GtsFile * f, 
+					   GtsFileVariable * vars,
+					   const gchar * name,
+					   const gchar * format,
+					   ...);
+void           gts_file_destroy           (GtsFile * f);
 
 /* Objects: object.c */
 
@@ -138,7 +204,8 @@ void gts_get_newline             (FILE * fptr);
 
 typedef enum
 {
-  GTS_DESTROYED         = 1 << 0
+  GTS_DESTROYED         = 1 << 0,
+  GTS_USER_FLAG         = 1 /* user flags start from here */
 } GtsObjectFlags;
 
 #define GTS_OBJECT_FLAGS(obj)             (GTS_OBJECT (obj)->flags)
@@ -169,7 +236,7 @@ struct _GtsObjectClass {
 
   void        (* clone)      (GtsObject *, GtsObject *);
   void        (* destroy)    (GtsObject *);
-  gint        (* read)       (GtsObject **, FILE *);
+  void        (* read)       (GtsObject **, GtsFile *);
   void        (* write)      (GtsObject *, FILE *);
   GtsColor    (* color)      (GtsObject *);
   void        (* attributes) (GtsObject *, GtsObject *);
@@ -244,12 +311,13 @@ void             gts_object_init                (GtsObject * object,
 						 GtsObjectClass * klass);
 void             gts_object_reset_reserved      (GtsObject * object);
 void             gts_object_destroy             (GtsObject * object);
+void             gts_finalize                   (void);
 
 /* Ranges: surface.c */
 typedef struct _GtsRange               GtsRange;
 
 struct _GtsRange {
-  gdouble min, max, mean, stddev;
+  gdouble min, max, sum, sum2, mean, stddev;
   guint n;
 };
 
@@ -280,6 +348,7 @@ struct _GtsPoint {
 
 struct _GtsPointClass {
   GtsObjectClass parent_class;
+  gboolean binary;
 };
 
 GtsPointClass * gts_point_class                      (void);
@@ -311,6 +380,10 @@ gdouble       gts_point_orientation_3d               (GtsPoint * p1,
 						      GtsPoint * p2,
 						      GtsPoint * p3,
 						      GtsPoint * p4);
+gint          gts_point_orientation_3d_sos           (GtsPoint * p1,
+						      GtsPoint * p2,
+						      GtsPoint * p3,
+						      GtsPoint * p4);
 GtsIntersect  gts_point_is_in_triangle               (GtsPoint * p,
 						      GtsTriangle * t);
 gdouble       gts_point_in_circle                    (GtsPoint * p, 
@@ -322,6 +395,9 @@ gdouble       gts_point_in_triangle_circle           (GtsPoint * p,
 gdouble       gts_point_orientation                  (GtsPoint * p1,
 						      GtsPoint * p2,
 						      GtsPoint * p3);
+gint          gts_point_orientation_sos              (GtsPoint * p1,
+						      GtsPoint * p2,
+						      GtsPoint * p3);
 gdouble       gts_point_segment_distance2            (GtsPoint * p, 
 						      GtsSegment * s);
 gdouble       gts_point_segment_distance             (GtsPoint * p, 
@@ -358,6 +434,10 @@ struct _GtsVertex {
 
 struct _GtsVertexClass {
   GtsPointClass parent_class;
+
+  void        (* intersection_attributes) (GtsVertex *, 
+					   GtsObject *, 
+					   GtsObject *);
 };
 
 GTS_C_VAR 
@@ -385,11 +465,52 @@ GSList *      gts_vertices_from_segments   (GSList * segments);
 gboolean      gts_vertex_is_boundary       (GtsVertex * v, 
 					    GtsSurface * surface);
 GList *       gts_vertices_merge           (GList * vertices, 
-					    gdouble epsilon);
+					    gdouble epsilon,
+					    gboolean (* check) (GtsVertex *, GtsVertex *));
 GSList *      gts_vertex_fan_oriented      (GtsVertex * v, 
 					    GtsSurface * surface);
 guint         gts_vertex_is_contact        (GtsVertex * v, gboolean sever);
 
+/* GtsVertexNormal: Header */
+
+typedef struct _GtsVertexNormal         GtsVertexNormal;
+
+struct _GtsVertexNormal {
+  /*< private >*/
+  GtsVertex parent;
+
+  /*< public >*/
+  GtsVector n;
+};
+
+#define GTS_VERTEX_NORMAL(obj)            GTS_OBJECT_CAST (obj,\
+					         GtsVertexNormal,\
+					         gts_vertex_normal_class ())
+#define GTS_IS_VERTEX_NORMAL(obj)         (gts_object_is_from_class (obj,\
+						 gts_vertex_normal_class ()))
+
+GtsVertexClass * gts_vertex_normal_class  (void);
+
+/* GtsColorVertex: Header */
+
+typedef struct _GtsColorVertex         GtsColorVertex;
+
+struct _GtsColorVertex {
+  /*< private >*/
+  GtsVertex parent;
+
+  /*< public >*/
+  GtsColor c;
+};
+
+#define GTS_COLOR_VERTEX(obj)            GTS_OBJECT_CAST (obj,\
+					         GtsColorVertex,\
+					         gts_color_vertex_class ())
+#define GTS_IS_COLOR_VERTEX(obj)         (gts_object_is_from_class (obj,\
+						 gts_color_vertex_class ()))
+
+GtsVertexClass * gts_color_vertex_class  (void);
+
 /* Segments: segment.c */
 
 #define GTS_IS_SEGMENT(obj) (gts_object_is_from_class (obj,\
@@ -485,6 +606,12 @@ GtsEdge *     gts_edge_is_duplicate               (GtsEdge * e);
 GList *       gts_edges_merge                     (GList * edges);
 gboolean      gts_edge_belongs_to_tetrahedron     (GtsEdge * e);
 guint         gts_edge_is_contact                 (GtsEdge * e);
+void          gts_edge_swap                       (GtsEdge * e, 
+						   GtsSurface * s);
+gboolean      gts_edge_manifold_faces             (GtsEdge * e, 
+						   GtsSurface * s,
+						   GtsFace ** f1, 
+						   GtsFace ** f2);
 
 /* Triangles: triangle.c */
 
@@ -630,17 +757,31 @@ void          gts_face_foreach_neighbor             (GtsFace * f,
 #define       gts_vector_scalar(v1, v2)    ((v1)[0]*(v2)[0] +\
 					    (v1)[1]*(v2)[1] +\
 					    (v1)[2]*(v2)[2])
-GtsMatrix *   gts_matrix_new          (gdouble a00, gdouble a01, gdouble a02,
-				       gdouble a10, gdouble a11, gdouble a12,
-				       gdouble a20, gdouble a21, gdouble a22);
-void          gts_matrix_assign       (GtsMatrix * m,
-				       gdouble a00, gdouble a01, gdouble a02,
-				       gdouble a10, gdouble a11, gdouble a12,
-				       gdouble a20, gdouble a21, gdouble a22);
+#define       gts_vector_norm(v)   (sqrt ((v)[0]*(v)[0] +\
+                                          (v)[1]*(v)[1] +\
+                                          (v)[2]*(v)[2]))
+#define       gts_vector_normalize(v) {\
+  gdouble __gts_n = gts_vector_norm (v);\
+  if (__gts_n > 0.) {\
+    (v)[0] /= __gts_n;\
+    (v)[1] /= __gts_n;\
+    (v)[2] /= __gts_n;\
+  }\
+}
+GtsMatrix * gts_matrix_new (gdouble a00, gdouble a01, gdouble a02, gdouble a03,
+			    gdouble a10, gdouble a11, gdouble a12, gdouble a13,
+			    gdouble a20, gdouble a21, gdouble a22, gdouble a23,
+			    gdouble a30, gdouble a31, gdouble a32, gdouble a33);
+void gts_matrix_assign (GtsMatrix * m,
+			gdouble a00, gdouble a01, gdouble a02, gdouble a03,
+			gdouble a10, gdouble a11, gdouble a12, gdouble a13,
+			gdouble a20, gdouble a21, gdouble a22, gdouble a23,
+			gdouble a30, gdouble a31, gdouble a32, gdouble a33);
 GtsMatrix *   gts_matrix_projection                  (GtsTriangle * t);
 GtsMatrix *   gts_matrix_transpose                   (GtsMatrix * m);
 gdouble       gts_matrix_determinant                 (GtsMatrix * m);
 GtsMatrix *   gts_matrix_inverse                     (GtsMatrix * m);
+GtsMatrix *   gts_matrix3_inverse                    (GtsMatrix * m);
 void          gts_matrix_print                       (GtsMatrix * m, 
 						      FILE * fptr);
 guint         gts_matrix_compatible_row              (GtsMatrix * A,
@@ -655,9 +796,20 @@ guint         gts_matrix_quadratic_optimization      (GtsMatrix * A,
 						      GtsVector c);
 GtsMatrix *   gts_matrix_product                     (GtsMatrix * m1, 
 						      GtsMatrix * m2);
+GtsMatrix *   gts_matrix_zero                        (GtsMatrix * m);
+GtsMatrix *   gts_matrix_identity                    (GtsMatrix * m);
+GtsMatrix *   gts_matrix_scale                       (GtsMatrix * m, 
+						      GtsVector s);
+GtsMatrix *   gts_matrix_translate                   (GtsMatrix * m, 
+						      GtsVector t);
+GtsMatrix *   gts_matrix_rotate                      (GtsMatrix * m,
+						      GtsVector r,
+						      gdouble angle);
 void          gts_matrix_destroy                     (GtsMatrix * m);
 void          gts_vector_print                       (GtsVector v,
 						      FILE * fptr);
+void          gts_vector4_print                      (GtsVector4 v, 
+						      FILE * fptr);
 
 /* Kdtrees: kdtree.c */
 
@@ -801,6 +953,9 @@ void       gts_bb_tree_destroy               (GNode * tree,
 
 typedef struct _GtsSurfaceStats        GtsSurfaceStats;
 typedef struct _GtsSurfaceQualityStats GtsSurfaceQualityStats;
+typedef GtsVertex * (*GtsRefineFunc)   (GtsEdge * e,
+					GtsVertexClass * klass,
+					gpointer data);
 typedef GtsVertex * (*GtsCoarsenFunc)  (GtsEdge * e,
 					GtsVertexClass * klass,
 					gpointer data);
@@ -830,7 +985,11 @@ struct _GtsSurfaceQualityStats {
 struct _GtsSurface {
   GtsObject object;
 
+#ifdef USE_SURFACE_BTREE
+  GTree * faces;
+#else /* not USE_SURFACE_BTREE */
   GHashTable * faces;
+#endif /* not USE_SURFACE_BTREE */
   GtsFaceClass * face_class;
   GtsEdgeClass * edge_class;
   GtsVertexClass * vertex_class;
@@ -863,7 +1022,7 @@ void         gts_surface_add_face          (GtsSurface * s,
 void         gts_surface_remove_face       (GtsSurface * s, 
 					    GtsFace * f);
 guint        gts_surface_read              (GtsSurface * surface,
-					    FILE * fptr);
+					    GtsFile * f);
 gdouble      gts_surface_area              (GtsSurface * s);
 void         gts_surface_stats             (GtsSurface * s, 
 					    GtsSurfaceStats * stats);
@@ -875,6 +1034,8 @@ void         gts_surface_write             (GtsSurface * s,
 					    FILE * fptr);
 void         gts_surface_write_oogl        (GtsSurface * s, 
 					    FILE * fptr);
+void         gts_surface_write_vtk         (GtsSurface * s, 
+					    FILE * fptr);
 void         gts_surface_write_oogl_boundary (GtsSurface * s, 
 					      FILE * fptr);
 void         gts_surface_foreach_vertex    (GtsSurface * s, 
@@ -896,9 +1057,12 @@ GtsFace *    gts_surface_traverse_next     (GtsSurfaceTraverse * t,
 					    guint * level);
 void         gts_surface_traverse_destroy  (GtsSurfaceTraverse * t);
 void         gts_surface_refine            (GtsSurface * surface,
-					    GCompareFunc compare_func,
+					    GtsKeyFunc cost_func,
+					    gpointer cost_data,
+					    GtsRefineFunc refine_func,
+					    gpointer refine_data,
 					    GtsStopFunc stop_func,
-					    gpointer data);
+					    gpointer stop_data);
 gboolean     gts_edge_collapse_is_valid    (GtsEdge * e);
 void         gts_surface_coarsen           (GtsSurface * surface,
 					    GtsKeyFunc cost_func,
@@ -914,6 +1078,11 @@ gboolean     gts_coarsen_stop_number       (gdouble cost,
 gboolean     gts_coarsen_stop_cost         (gdouble cost, 
 					    guint nedge, 
 					    gdouble * max_cost);
+void         gts_surface_tessellate        (GtsSurface * s,
+					    GtsRefineFunc refine_func,
+					    gpointer refine_data);
+GtsSurface * gts_surface_generate_sphere   (GtsSurface * s,
+					    guint geodesation_order);
 GtsSurface * gts_surface_copy              (GtsSurface * s1,
 					    GtsSurface * s2);
 void         gts_surface_merge             (GtsSurface * s, 
@@ -922,6 +1091,10 @@ gboolean     gts_surface_is_manifold       (GtsSurface * s);
 gboolean     gts_surface_is_closed         (GtsSurface * s);
 gboolean     gts_surface_is_orientable     (GtsSurface * s);
 gdouble      gts_surface_volume            (GtsSurface * s);
+gdouble      gts_surface_center_of_mass    (GtsSurface * s,
+					    GtsVector cm);
+gdouble      gts_surface_center_of_area    (GtsSurface * s,
+					    GtsVector cm);
 guint        gts_surface_vertex_number     (GtsSurface * s);
 guint        gts_surface_edge_number       (GtsSurface * s);
 guint        gts_surface_face_number       (GtsSurface * s);
@@ -931,6 +1104,26 @@ void         gts_surface_distance          (GtsSurface * s1,
 					    GtsRange * face_range, 
 					    GtsRange * boundary_range);
 GSList *     gts_surface_boundary          (GtsSurface * surface);
+GSList *     gts_surface_split             (GtsSurface * s);
+
+/* Discrete differential operators: curvature.c */
+
+gboolean gts_vertex_mean_curvature_normal  (GtsVertex * v, 
+					    GtsSurface * s, 
+					    GtsVector Kh);
+gboolean gts_vertex_gaussian_curvature     (GtsVertex * v, 
+					    GtsSurface * s, 
+					    gdouble * Kg);
+void     gts_vertex_principal_curvatures   (gdouble Kh, 
+					    gdouble Kg, 
+					    gdouble * K1, 
+					    gdouble * K2);
+void     gts_vertex_principal_directions   (GtsVertex * v, 
+					    GtsSurface * s, 
+					    GtsVector Kh,
+                                            gdouble Kg,
+					    GtsVector e1, 
+					    GtsVector e2);
 
 /* Volume optimization: vopt.c */
 typedef struct _GtsVolumeOptimizedParams   GtsVolumeOptimizedParams;
@@ -949,6 +1142,11 @@ gdouble      gts_volume_optimized_cost     (GtsEdge * e,
 
 /* Boolean operations: boolean.c */
 
+GSList *     gts_surface_intersection      (GtsSurface * s1,
+					    GtsSurface * s2,
+					    GNode * faces_tree1,
+					    GNode * faces_tree2);
+
 typedef struct _GtsSurfaceInter         GtsSurfaceInter;
 typedef struct _GtsSurfaceInterClass    GtsSurfaceInterClass;
 typedef enum { GTS_1_OUT_2, 
@@ -984,13 +1182,21 @@ gts_surface_inter_new            (GtsSurfaceInterClass * klass,
 				  GtsSurface * s1,
 				  GtsSurface * s2,
 				  GNode * faces_tree1,
-				  GNode * faces_tree2);
+				  GNode * faces_tree2,
+				  gboolean is_open1,
+				  gboolean is_open2);
 gboolean 
-gts_surface_inter_check          (GtsSurfaceInter * si);
+gts_surface_inter_check          (GtsSurfaceInter * si,
+				  gboolean * closed);
 void 
 gts_surface_inter_boolean        (GtsSurfaceInter * si, 
 				  GtsSurface * surface,
 				  GtsBooleanOperation op);
+gboolean gts_surface_foreach_intersecting_face (GtsSurface * s,
+					    GtsBBTreeTraverseFunc func,
+					    gpointer data);
+GtsSurface * 
+gts_surface_is_self_intersecting (GtsSurface * s);
 
 /* Binary Heap: heap.c */
 
@@ -1064,6 +1270,7 @@ gboolean       gts_fifo_is_empty      (GtsFifo * fifo);
 void           gts_fifo_foreach       (GtsFifo * fifo, 
 				       GtsFunc func, 
 				       gpointer data);
+void           gts_fifo_reverse       (GtsFifo * fifo);
 void           gts_fifo_destroy       (GtsFifo * fifo);
 
 /* Progressive surfaces */
@@ -1177,10 +1384,9 @@ void          gts_psurface_write              (GtsPSurface * ps,
 GtsPSurface * gts_psurface_open               (GtsPSurfaceClass * klass,
 					       GtsSurface * s,
 					       GtsSplitClass * split_class,
-					       FILE * fptr,
-					       guint * line);
+					       GtsFile * f);
 GtsSplit *    gts_psurface_read_vertex        (GtsPSurface * ps, 
-					       FILE * fptr);
+					       GtsFile * fp);
 void          gts_psurface_close              (GtsPSurface * ps);
 void          gts_psurface_foreach_vertex     (GtsPSurface * ps, 
 					       GtsFunc func, 
@@ -1299,6 +1505,26 @@ GSList *             gts_delaunay_add_constraint (GtsSurface * surface,
 						  GtsConstraint * c);
 void                 gts_delaunay_remove_hull    (GtsSurface * surface);
 
+/* GtsListFace: Header */
+
+typedef struct _GtsListFace         GtsListFace;
+
+struct _GtsListFace {
+  /*< private >*/
+  GtsFace parent;
+
+  /*< public >*/
+  GSList * points;
+};
+
+#define GTS_LIST_FACE(obj)            GTS_OBJECT_CAST (obj,\
+					         GtsListFace,\
+					         gts_list_face_class ())
+#define GTS_IS_LIST_FACE(obj)         (gts_object_is_from_class (obj,\
+						 gts_list_face_class ()))
+
+GtsFaceClass *       gts_list_face_class         (void);
+
 /* Constrained Delaunay refinement: refine.c */
 
 typedef gboolean   (* GtsEncroachFunc)           (GtsVertex * v,
@@ -1323,7 +1549,7 @@ guint                gts_delaunay_refine         (GtsSurface * surface,
 						  GtsKeyFunc cost,
 						  gpointer cost_data);
 
-/* Isosurfaces: iso.c */
+/* Isosurfaces (marching cubes): iso.c */
 
 typedef struct _GtsGridPlane     GtsGridPlane;
 typedef struct _GtsIsoSlice      GtsIsoSlice;
@@ -1371,6 +1597,24 @@ void           gts_isosurface_cartesian     (GtsSurface * surface,
 					     gpointer data,
 					     gdouble iso);
 
+/* Isosurfaces (marching tetrahedra): isotetra.c */
+
+void           gts_isosurface_tetra         (GtsSurface * surface,
+					     GtsCartesianGrid g,
+					     GtsIsoCartesianFunc f,
+					     gpointer data,
+					     gdouble iso);
+void           gts_isosurface_tetra_bcl     (GtsSurface * surface,
+					     GtsCartesianGrid g,
+					     GtsIsoCartesianFunc f,
+					     gpointer data,
+					     gdouble iso);
+void           gts_isosurface_tetra_bounded (GtsSurface * surface,
+					     GtsCartesianGrid g,
+					     GtsIsoCartesianFunc f,
+					     gpointer data,
+					     gdouble iso);
+
 /* Named vertices, edges and triangles: named.c */
 
 #define GTS_NAME_LENGTH             40
@@ -1548,7 +1792,9 @@ struct _GtsContaineeClass {
 
   void     (* add_container)    (GtsContainee *, GtsContainer *);
   void     (* remove_container) (GtsContainee *, GtsContainer *);
+  void     (* foreach)          (GtsContainee *, GtsFunc, gpointer);
   gboolean (* is_contained)     (GtsContainee *, GtsContainer *);
+  void     (* replace)          (GtsContainee *, GtsContainee *);
 };
 
 #define GTS_CONTAINEE(obj)            GTS_OBJECT_CAST (obj,\
@@ -1564,6 +1810,8 @@ GtsContaineeClass * gts_containee_class        (void);
 GtsContainee *      gts_containee_new          (GtsContaineeClass * klass);
 gboolean            gts_containee_is_contained (GtsContainee * item, 
 						GtsContainer * c);
+void                gts_containee_replace      (GtsContainee * item,
+						GtsContainee * with);
 
 /* GtsSListContainee: container.c */
 
@@ -1697,6 +1945,7 @@ struct _GtsGNodeClass {
   GtsSListContainerClass parent_class;
 
   gfloat (* weight) (GtsGNode *);
+  void   (* write)  (GtsGNode *, FILE *);
 };
 
 #define GTS_GNODE(obj)            GTS_OBJECT_CAST (obj,\
@@ -1785,6 +2034,34 @@ GtsWGNodeClass * gts_wgnode_class                (void);
 GtsWGNode *      gts_wgnode_new                  (GtsWGNodeClass * klass,
 						  gfloat weight);
 
+/* GtsPNode */
+
+typedef struct _GtsPNode         GtsPNode;
+typedef struct _GtsPNodeClass    GtsPNodeClass;
+
+struct _GtsPNode {
+  GtsGNode node;
+
+  gpointer data;
+};
+
+struct _GtsPNodeClass {
+  GtsGNodeClass parent_class;
+};
+
+#define GTS_PNODE(obj)            GTS_OBJECT_CAST (obj,\
+					           GtsPNode,\
+					           gts_pnode_class ())
+#define GTS_PNODE_CLASS(klass)    GTS_OBJECT_CLASS_CAST (klass,\
+						         GtsPNodeClass,\
+						         gts_pnode_class())
+#define GTS_IS_PNODE(obj)         (gts_object_is_from_class (obj,\
+						   gts_pnode_class ()))
+     
+GtsPNodeClass * gts_pnode_class                (void);
+GtsPNode *      gts_pnode_new                  (GtsPNodeClass * klass,
+						gpointer data);
+
 /* GtsFNode */
 
 typedef struct _GtsFNode         GtsFNode;
@@ -1828,7 +2105,9 @@ struct _GtsGEdge {
 struct _GtsGEdgeClass {
   GtsContaineeClass parent_class;
 
-  gfloat (* weight) (GtsGEdge *);
+  GtsGEdge * (* link)   (GtsGEdge * e, GtsGNode * n1, GtsGNode * n2);
+  gfloat     (* weight) (GtsGEdge * e);
+  void       (* write)  (GtsGEdge * e, FILE * fp);
 };
 
 #define GTS_GEDGE(obj)            GTS_OBJECT_CAST (obj,\
@@ -1848,6 +2127,36 @@ gfloat          gts_gedge_weight               (GtsGEdge * e);
 #define         gts_gedge_connects(e, a1, a2)\
    (((e)->n1 == a1 && (e)->n2 == a2) || ((e)->n1 == a2 && (e)->n2 == a1)) 
 
+/* GtsPGEdge: graph.c */
+
+typedef struct _GtsPGEdge         GtsPGEdge;
+typedef struct _GtsPGEdgeClass    GtsPGEdgeClass;
+
+struct _GtsPGEdge {
+  GtsGEdge gedge;
+
+  gpointer data;
+};
+
+struct _GtsPGEdgeClass {
+  GtsGEdgeClass parent_class;
+};
+
+#define GTS_PGEDGE(obj)            GTS_OBJECT_CAST (obj,\
+					           GtsPGEdge,\
+					           gts_pgedge_class ())
+#define GTS_PGEDGE_CLASS(klass)    GTS_OBJECT_CLASS_CAST (klass,\
+						         GtsPGEdgeClass,\
+						         gts_pgedge_class())
+#define GTS_IS_PGEDGE(obj)         (gts_object_is_from_class (obj,\
+						   gts_pgedge_class ()))
+     
+GtsPGEdgeClass * gts_pgedge_class                (void);
+GtsPGEdge *      gts_pgedge_new                  (GtsPGEdgeClass * klass,
+						  GtsGNode * n1,
+						  GtsGNode * n2,
+						  gpointer data);
+
 /* GtsWGEdge: graph.c */
 
 typedef struct _GtsWGEdge         GtsWGEdge;
@@ -1883,6 +2192,7 @@ GtsWGEdge *      gts_wgedge_new                  (GtsWGEdgeClass * klass,
 struct _GtsGraph {
   GtsHashContainer object;
 
+  GtsGraphClass * graph_class;
   GtsGNodeClass * node_class;
   GtsGEdgeClass * edge_class;
 };
@@ -1928,8 +2238,13 @@ GtsGNode *         gts_graph_farthest            (GtsGraph * g,
 						  GSList * gnodes);
 guint              gts_graph_edges_cut           (GtsGraph * g);
 gfloat             gts_graph_edges_cut_weight    (GtsGraph * g);
-guint              gts_graph_read                (GtsGraph * g, 
+void               gts_graph_write               (GtsGraph * g, 
 						  FILE * fp);
+void               gts_graph_write_dot           (GtsGraph * g, 
+						  FILE * fp);
+GtsGraph *         gts_graph_read                (GtsFile * fp);
+guint              gts_graph_read_jostle         (GtsGraph * g, 
+						  GtsFile * fp);
 
 /* GtsWGraph: graph.c */
 
@@ -1965,6 +2280,11 @@ GtsGraph *       gts_surface_graph_new           (GtsGraphClass * klass,
 GtsSurface *     gts_surface_graph_surface       (GtsGraph * surface_graph,
 						  GtsSurface * s);
 
+/* Segments graph: graph.c */
+
+GtsGraph *       gts_segments_graph_new          (GtsGraphClass * klass,
+						  GSList * segments);
+
 /* GtsGNodeSplit: pgraph.c */
 
 typedef struct _GtsGNodeSplit         GtsGNodeSplit;
@@ -2070,7 +2390,8 @@ GSList *         gts_graph_recursive_bisection        (GtsWGraph * wg,
 						       guint n,
 						       guint ntry,
 						       guint mmax,
-						       guint nmin);
+						       guint nmin,
+						       gfloat imbalance);
 void             gts_graph_partition_destroy          (GSList * partition);
 
 /* Graph bisection: partition.c */
@@ -2093,11 +2414,13 @@ GtsGraphBisection * gts_graph_bfgg_bisection       (GtsGraph * g,
 gdouble             gts_graph_bisection_kl_refine  (GtsGraphBisection * bg,
 						    guint mmax);
 gdouble             gts_graph_bisection_bkl_refine (GtsGraphBisection * bg,
-						    guint mmax);
+						    guint mmax,
+						    gfloat imbalance);
 GtsGraphBisection * gts_graph_bisection_new        (GtsWGraph * wg,
 						    guint ntry,
 						    guint mmax,
-						    guint nmin);
+						    guint nmin,
+						    gfloat imbalance);
 void                gts_graph_bisection_destroy    (GtsGraphBisection * bg,
 						    gboolean destroy_graphs);
 
