@@ -136,10 +136,19 @@ void gts_file_error (GtsFile * f,
   va_end (args);
 }
 
-static gint file_getc (GtsFile * f)
+/**
+ * gts_file_getc :
+ * @f: a #GtsFile.
+ *
+ * Returns: the next character in @f or EOF if the end of the file is
+ * reached or if an error occured.
+ */
+gint gts_file_getc (GtsFile * f)
 {
   gint c;
 
+  g_return_val_if_fail (f != NULL, EOF);
+
   if (f->type == GTS_ERROR)
     return EOF;
 
@@ -165,19 +174,63 @@ static gint file_getc (GtsFile * f)
   return c;
 }
 
-static gint file_getc_scope (GtsFile * f)
+/**
+ * gts_file_read:
+ * @f: a #GtsFile.
+ * @ptr: a pointer.
+ * @size: size of an element.
+ * @nmemb: number of elements.
+ *
+ * Reads @nmemb elements of data, each @size bytes long, from @f,
+ * storing them at the location given by @ptr.
+ *
+ * Returns: the number of elements read.
+ */
+guint gts_file_read (GtsFile * f, gpointer ptr, guint size, guint nmemb)
+{
+  guint i, n;
+  gchar * p;
+
+  g_return_val_if_fail (f != NULL, 0);
+  g_return_val_if_fail (ptr != NULL, 0);
+
+  if (f->type == GTS_ERROR)
+    return 0;
+
+  n = fread (ptr, size, nmemb, f->fp);
+  for (i = 0, p = ptr; i < n*size; i++, p++) {
+    f->curpos++;
+    if (*p == '\n') {
+      f->curline++;
+      f->curpos = 0;
+    }
+  }
+  return n;
+}
+
+/**
+ * gts_file_getc_scope :
+ * @f: a #GtsFile.
+ *
+ * Returns: the next character in @f in the scope defined by
+ * @f->scope_max or EOF if the end of the file is reached or if an
+ * error occured.
+ */
+gint gts_file_getc_scope (GtsFile * f)
 {
   gint c;
 
+  g_return_val_if_fail (f != NULL, EOF);
+
   if (f->type == GTS_ERROR)
     return EOF;
   
   if (f->scope <= f->scope_max)
-    c = file_getc (f);
+    c = gts_file_getc (f);
   else {
-    c = file_getc (f);
+    c = gts_file_getc (f);
     while (c != EOF && f->scope > f->scope_max)
-      c = file_getc (f);    
+      c = gts_file_getc (f);    
   }
   return c;
 }
@@ -186,9 +239,9 @@ static void jump_to (GtsFile * f, gchar c)
 {
   gint a;
   
-  a = file_getc_scope (f);
+  a = gts_file_getc_scope (f);
   while (a != EOF && a != c)
-    a = file_getc_scope (f);
+    a = gts_file_getc_scope (f);
 }
 
 /**
@@ -217,7 +270,7 @@ void gts_file_next_token (GtsFile * f)
     return;
   }
   f->type = GTS_NONE;
-  c = file_getc_scope (f);
+  c = gts_file_getc_scope (f);
   while (c != EOF && (!in_string || !char_in_string (c, f->delimiters))) {
     if (char_in_string (c, f->comments))
       jump_to (f, '\n');
@@ -238,7 +291,7 @@ void gts_file_next_token (GtsFile * f)
 	break;
       }
     }
-    c = file_getc_scope (f);
+    c = gts_file_getc_scope (f);
   }
   if (f->type == GTS_NONE && f->token->len > 0) {
     gchar * a;
