@@ -17,6 +17,8 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#include <math.h>
+
 #include "gts.h"
 
 /* #define DEBUG */
@@ -494,6 +496,7 @@ GtsGraphBisection * gts_graph_ggg_bisection (GtsGraph * g, guint ntry)
 {
   gfloat size, bestcost = G_MAXFLOAT, smin;
   GtsGraph * bestg1 = NULL, * bestg2 = NULL;
+  gboolean balanced = FALSE;
   GtsEHeap * degree_heap;
   GtsGNode * seed;
   GtsGraphBisection * bg;
@@ -544,7 +547,9 @@ GtsGraphBisection * gts_graph_ggg_bisection (GtsGraph * g, guint ntry)
     gts_container_foreach (GTS_CONTAINER (g), (GtsFunc) add_unused, g2);
 
     cost = gts_graph_edges_cut_weight (g1);
-    if (!bestg1 || (cost < bestcost && gts_graph_weight (g1) >= smin)) {
+    if (!bestg1 || 
+	(!balanced && gts_graph_weight (g1) >= smin) ||
+	(cost < bestcost && gts_graph_weight (g1) >= smin)) {
       if (bestg1)
 	bestcost = cost;
       if (bestg1)
@@ -553,6 +558,8 @@ GtsGraphBisection * gts_graph_ggg_bisection (GtsGraph * g, guint ntry)
 	gts_object_destroy (GTS_OBJECT (bestg2));
       bestg1 = g1;
       bestg2 = g2;
+      if (gts_graph_weight (g1) >= smin)
+	balanced = TRUE;
     }
     else {
       gts_object_destroy (GTS_OBJECT (g1));
@@ -715,7 +722,7 @@ gdouble gts_graph_bisection_kl_refine (GtsGraphBisection * bg,
   GtsGNode * n;
   guint nm = 0, i;
   GtsGNode ** moves;
-  gdouble bestcost = 0., totalcost = 0.;
+  gdouble bestcost = 0., totalcost = 0., best_balance;
 
   g_return_val_if_fail (bg != NULL, 0.);
   g_return_val_if_fail (mmax > 0, 0.);
@@ -731,6 +738,7 @@ gdouble gts_graph_bisection_kl_refine (GtsGraphBisection * bg,
   gts_eheap_thaw (h2);
 
   moves = g_malloc (sizeof (GtsGNode *)*mmax);
+  best_balance = fabs (gts_graph_weight (bg->g1) - gts_graph_weight (bg->g2));
 
   do {
     GtsGraph * g1, * g2;
@@ -758,6 +766,14 @@ gdouble gts_graph_bisection_kl_refine (GtsGraphBisection * bg,
 	bestcost = totalcost;
 	nm = 0;
       }
+      else if (totalcost == bestcost) {
+	gdouble balance = fabs (gts_graph_weight (g1) - gts_graph_weight (g2));
+
+	if (balance < best_balance) {
+	  best_balance = balance;
+	  nm = 0;
+	}
+      }	       
       else
 	moves[nm++] = n;
 
@@ -850,6 +866,8 @@ static void update_neighbors (GtsGNode * n, GtsGraphBisection * bg,
  * gts_graph_bisection_bkl_refine:
  * @bg: a #GtsGraphBisection.
  * @mmax: the maximum number of unsuccessful successive moves.
+ * @imbalance: the maximum relative imbalance allowed between the
+ * weights of both halves of the partition.
  *
  * An implementation of the simplified boundary Kernighan-Lin
  * algorithm for graph bisection refinement as described in Karypis
@@ -862,16 +880,19 @@ static void update_neighbors (GtsGNode * n, GtsGraphBisection * bg,
  * Returns: the decrease in the weight of the edges cut by the bisection.  
  */
 gdouble gts_graph_bisection_bkl_refine (GtsGraphBisection * bg,
-					guint mmax)
+					guint mmax,
+					gfloat imbalance)
 {
   GtsEHeap * h1, * h2;
   GtsGNode * n;
   guint nm = 0, i;
   GtsGNode ** moves;
-  gdouble bestcost = 0., totalcost = 0.;
+  gdouble bestcost = 0., totalcost = 0., best_balance;
+  gboolean balanced = FALSE;
 
   g_return_val_if_fail (bg != NULL, 0.);
   g_return_val_if_fail (mmax > 0, 0.);
+  g_return_val_if_fail (imbalance >= 0. && imbalance <= 1., 0.);
 
   h1 = gts_eheap_new ((GtsKeyFunc) node_move_cost1, bg);
   gts_eheap_freeze (h1);
@@ -884,6 +905,10 @@ gdouble gts_graph_bisection_bkl_refine (GtsGraphBisection * bg,
   gts_eheap_thaw (h2);
 
   moves = g_malloc (sizeof (GtsGNode *)*mmax);
+  imbalance *= gts_graph_weight (bg->g);
+  best_balance = fabs (gts_graph_weight (bg->g1) - gts_graph_weight (bg->g2));
+  if (best_balance <= imbalance)
+    balanced = TRUE;
 
   do {
     GtsGraph * g1, * g2;
@@ -905,6 +930,8 @@ gdouble gts_graph_bisection_bkl_refine (GtsGraphBisection * bg,
       bg2 = bg->bg1;
     }
     if (n) {
+      gdouble balance;
+	
       GTS_OBJECT (n)->reserved = n;
       gts_container_add (GTS_CONTAINER (g2), GTS_CONTAINEE (n));
       gts_container_remove (GTS_CONTAINER (g1), GTS_CONTAINEE (n));
@@ -913,10 +940,24 @@ gdouble gts_graph_bisection_bkl_refine (GtsGraphBisection * bg,
 	g_hash_table_insert (bg2, n, n);
 
       update_neighbors (n, bg, h1, h2);
-  
+
       totalcost += cost;
-      if (totalcost < bestcost) {
+      balance = fabs (gts_graph_weight (g1) - gts_graph_weight (g2));
+      
+      if (!balanced && balance <= imbalance) {
 	bestcost = totalcost;
+	best_balance = balance;
+	balanced = TRUE;
+	nm = 0;
+      }
+      else if (totalcost < bestcost && 
+	       (balance < best_balance || balance <= imbalance)) {
+	bestcost = totalcost;
+	best_balance = balance;
+	nm = 0;
+      }
+      else if (totalcost == bestcost && balance < best_balance) {
+	best_balance = balance;
 	nm = 0;
       }
       else
@@ -1004,6 +1045,8 @@ static void bisection_children (GtsGNodeSplit * ns, GtsGraphBisection * bg)
  * @ntry: the number of tries for the graph growing algorithm.
  * @mmax: the number of unsucessful moves for the refinement algorithm.
  * @nmin: the minimum number of nodes of the coarsest graph.
+ * @imbalance: the maximum relative imbalance allowed between the
+ * weights of both halves of the partition.
  *
  * An implementation of a multilevel bisection algorithm as presented
  * in Karypis and Kumar (1997). A multilevel hierarchy of graphs is
@@ -1017,7 +1060,8 @@ static void bisection_children (GtsGNodeSplit * ns, GtsGraphBisection * bg)
 GtsGraphBisection * gts_graph_bisection_new (GtsWGraph * wg,
 					     guint ntry,
 					     guint mmax,
-					     guint nmin)
+					     guint nmin,
+					     gfloat imbalance)
 {
   GtsGraph * g;
   GtsPGraph * pg;
@@ -1042,7 +1086,7 @@ GtsGraphBisection * gts_graph_bisection_new (GtsWGraph * wg,
 	   gts_graph_edges_cut_weight (bg->g1));
   g_assert (gts_graph_bisection_check (bg));
 #endif
-  while ((cost = gts_graph_bisection_bkl_refine (bg, mmax))) {
+  while ((cost = gts_graph_bisection_bkl_refine (bg, mmax, imbalance))) {
 #ifdef DEBUG
     fprintf (stderr, "  cost: %g\n", cost);
     g_assert (gts_graph_bisection_check (bg));
@@ -1063,7 +1107,7 @@ GtsGraphBisection * gts_graph_bisection_new (GtsWGraph * wg,
 	     gts_graph_edges_cut (bg->g1),
 	     gts_graph_edges_cut_weight (bg->g1));	   
 #endif
-    while ((cost = gts_graph_bisection_bkl_refine (bg, mmax))) {
+    while ((cost = gts_graph_bisection_bkl_refine (bg, mmax, imbalance))) {
 #ifdef DEBUG
       fprintf (stderr, "  cost: %g\n", cost);
       g_assert (gts_graph_bisection_check (bg));
@@ -1111,19 +1155,23 @@ static void recursive_bisection (GtsWGraph * wg,
 				 guint ntry,
 				 guint mmax,
 				 guint nmin,
+				 gfloat imbalance,
 				 GSList ** list)
 {
   if (np == 0)
     *list = g_slist_prepend (*list, wg);
   else {
-    GtsGraphBisection * bg = gts_graph_bisection_new (wg, ntry, mmax, nmin);
+    GtsGraphBisection * bg = 
+      gts_graph_bisection_new (wg, ntry, mmax, nmin, imbalance);
     GtsGraph * g1 = bg->g1;
     GtsGraph * g2 = bg->g2;
 
     gts_object_destroy (GTS_OBJECT (wg));
     gts_graph_bisection_destroy (bg, FALSE);
-    recursive_bisection (GTS_WGRAPH (g1), np - 1, ntry, mmax, nmin, list);
-    recursive_bisection (GTS_WGRAPH (g2), np - 1, ntry, mmax, nmin, list);
+    recursive_bisection (GTS_WGRAPH (g1), np - 1, ntry, mmax, nmin, imbalance,
+			 list);
+    recursive_bisection (GTS_WGRAPH (g2), np - 1, ntry, mmax, nmin, imbalance,
+			 list);
   }
 }
 
@@ -1134,6 +1182,8 @@ static void recursive_bisection (GtsWGraph * wg,
  * @ntry: the number of tries for the graph growing algorithm.
  * @mmax: the number of unsucessful moves for the refinement algorithm.
  * @nmin: the minimum number of nodes of the coarsest graph.
+ * @imbalance: the maximum relative imbalance allowed between the
+ * weights of both halves of the partition.
  *
  * Calls gts_graph_bisection_new() recursively in order to obtain a
  * 2^@n partition of @wg.
@@ -1144,7 +1194,8 @@ GSList * gts_graph_recursive_bisection (GtsWGraph * wg,
 					guint n,
 					guint ntry,
 					guint mmax,
-					guint nmin)
+					guint nmin,
+					gfloat imbalance)
 {
   GtsGraphBisection * bg;
   GtsGraph * g1, * g2;
@@ -1153,12 +1204,14 @@ GSList * gts_graph_recursive_bisection (GtsWGraph * wg,
   g_return_val_if_fail (wg != NULL, NULL);
   g_return_val_if_fail (n > 0, NULL);
   
-  bg = gts_graph_bisection_new (wg, ntry, mmax, nmin);
+  bg = gts_graph_bisection_new (wg, ntry, mmax, nmin, imbalance);
   g1 = bg->g1;
   g2 = bg->g2;
   gts_graph_bisection_destroy (bg, FALSE);
-  recursive_bisection (GTS_WGRAPH (g1), n - 1, ntry, mmax, nmin, &list);
-  recursive_bisection (GTS_WGRAPH (g2), n - 1, ntry, mmax, nmin, &list);
+  recursive_bisection (GTS_WGRAPH (g1), n - 1, ntry, mmax, nmin, imbalance, 
+		       &list);
+  recursive_bisection (GTS_WGRAPH (g2), n - 1, ntry, mmax, nmin, imbalance,
+		       &list);
 
   return list;
 }
