@@ -29,34 +29,56 @@ static void point_read (GtsObject ** o, GtsFile * f)
 {
   GtsPoint * p = GTS_POINT (*o);
 
-  if (f->type != GTS_INT && f->type != GTS_FLOAT) {
-    gts_file_error (f, "expecting a number (x coordinate)");
-    return;
+  if (GTS_POINT_CLASS ((*o)->klass)->binary) {
+    if (gts_file_read (f, &(p->x), sizeof (gdouble), 1) != 1) {
+      gts_file_error (f, "expecting a binary number (x coordinate)");
+      return;
+    }
+    if (gts_file_read (f, &(p->y), sizeof (gdouble), 1) != 1) {
+      gts_file_error (f, "expecting a binary number (y coordinate)");
+      return;
+    }
+    if (gts_file_read (f, &(p->z), sizeof (gdouble), 1) != 1) {
+      gts_file_error (f, "expecting a binary number (z coordinate)");
+      return;
+    }
   }
-  p->x = atof (f->token->str);
-
-  gts_file_next_token (f);
-  if (f->type != GTS_INT && f->type != GTS_FLOAT) {
-    gts_file_error (f, "expecting a number (y coordinate)");
-    return;
+  else {
+    if (f->type != GTS_INT && f->type != GTS_FLOAT) {
+      gts_file_error (f, "expecting a number (x coordinate)");
+      return;
+    }
+    p->x = atof (f->token->str);
+    
+    gts_file_next_token (f);
+    if (f->type != GTS_INT && f->type != GTS_FLOAT) {
+      gts_file_error (f, "expecting a number (y coordinate)");
+      return;
+    }
+    p->y = atof (f->token->str);
+    
+    gts_file_next_token (f);
+    if (f->type != GTS_INT && f->type != GTS_FLOAT) {
+      gts_file_error (f, "expecting a number (z coordinate)");
+      return;
+    }
+    p->z = atof (f->token->str);
+    
+    gts_file_next_token (f);
   }
-  p->y = atof (f->token->str);
-
-  gts_file_next_token (f);
-  if (f->type != GTS_INT && f->type != GTS_FLOAT) {
-    gts_file_error (f, "expecting a number (z coordinate)");
-    return;
-  }
-  p->z = atof (f->token->str);
-
-  gts_file_next_token (f);
 }
 
 static void point_write (GtsObject * o, FILE * fptr)
 {
   GtsPoint * p = GTS_POINT (o);
 
-  fprintf (fptr, "%.10g %.10g %.10g", p->x, p->y, p->z);
+  if (GTS_POINT_CLASS ((o)->klass)->binary) {
+    fwrite (&(p->x), sizeof (gdouble), 1, fptr);
+    fwrite (&(p->y), sizeof (gdouble), 1, fptr);
+    fwrite (&(p->z), sizeof (gdouble), 1, fptr);
+  }
+  else
+    fprintf (fptr, "%.10g %.10g %.10g", p->x, p->y, p->z);
 }
 
 static void point_class_init (GtsObjectClass * klass)
@@ -711,9 +733,9 @@ void gts_point_transform (GtsPoint * p, GtsMatrix * m)
 {
   gdouble x, y, z;
   g_return_if_fail (p != NULL && m != NULL);
-  x = m[0][0]*p->x + m[0][1]*p->y + m[0][2]*p->z;
-  y = m[1][0]*p->x + m[1][1]*p->y + m[1][2]*p->z;
-  z = m[2][0]*p->x + m[2][1]*p->y + m[2][2]*p->z;
+  x = m[0][0]*p->x + m[0][1]*p->y + m[0][2]*p->z + m[0][3];
+  y = m[1][0]*p->x + m[1][1]*p->y + m[1][2]*p->z + m[1][3];
+  z = m[2][0]*p->x + m[2][1]*p->y + m[2][2]*p->z + m[2][3];
   p->x = x; p->y = y; p->z = z;
 }
 
@@ -811,8 +833,7 @@ gboolean gts_point_is_inside_surface (GtsPoint * p,
   g_return_val_if_fail (tree != NULL, FALSE);
 
   bb = tree->data;
-  p1 = gts_point_new (gts_point_class (), p->x, p->y, 
-		      bb->z2 + fabs (bb->z2)/10.);
+  p1 = gts_point_new (gts_point_class (), bb->x2 + fabs (bb->x2)/10., p->y, p->z);
   i = list = gts_bb_tree_stabbed (tree, p);
   while (i) {
     GtsTriangle * t = GTS_TRIANGLE (GTS_BBOX (i->data)->bounded);
@@ -985,3 +1006,65 @@ gint gts_point_orientation_3d_sos (GtsPoint * p1,
   }
 #endif /* USE_ROBUST_PREDICATES */
 }
+
+/**
+ * gts_point_orientation_sos:
+ * @p1: a #GtsPoint.
+ * @p2: a #GtsPoint.
+ * @p3: a #GtsPoint.
+ *
+ * Checks for orientation of the projection of three points on the
+ * (x,y) plane.
+ *
+ * Simulation of Simplicity (SoS) is used to break ties when the
+ * orientation is degenerate (i.e. @p3 lies on the line defined by
+ * @p1 and @p2).
+ *
+ * Returns: a positive value if @p1, @p2 and @p3 appear in
+ * counterclockwise order or a negative value if they appear in
+ * clockwise order.  
+ */
+gint gts_point_orientation_sos (GtsPoint * p1,
+				GtsPoint * p2,
+				GtsPoint * p3)
+{
+#ifndef USE_ROBUST_PREDICATES
+  g_assert_not_reached ();
+  return 0;
+#else /* USE_ROBUST_PREDICATES */
+  gdouble o;
+
+  g_return_val_if_fail (p1 != NULL && p2 != NULL && p3 != NULL, 0);
+
+  o = orient2d ((gdouble *) &p1->x, 
+		(gdouble *) &p2->x, 
+		(gdouble *) &p3->x);
+  if (o != 0.)
+    return SIGN (o);
+  else {
+    GtsPoint * p[3];
+    gint sign;
+
+    p[0] = p1; p[1] = p2; p[2] = p3;
+    sign = sortp ((gpointer *) p, 3);
+    
+    /* epsilon^1/4 */
+    o = ORIENT1D (p[1]->x, p[2]->x);
+    if (o != 0.)
+      return - SIGN (o)*sign;
+    
+    /* epsilon^1/2 */
+    o = ORIENT1D (p[1]->y, p[2]->y);
+    if (o != 0.)
+      return SIGN (o)*sign;
+    
+    /* epsilon */
+    o = ORIENT1D (p[0]->x, p[2]->x);
+    if (o != 0.)
+      return SIGN (o)*sign;
+    
+    /* epsilon^3/2 */
+    return sign;
+  }
+#endif /* USE_ROBUST_PREDICATES */
+}
