@@ -19,6 +19,7 @@
 
 #include <stdlib.h>
 #include <math.h>
+#include <string.h>
 #include "gts.h"
 
 #include "gts-private.h"
@@ -51,6 +52,8 @@ static void surface_write (GtsObject * object, FILE * fptr)
 	   object->klass->info.name,
 	   GTS_OBJECT_CLASS (GTS_SURFACE (object)->face_class)->info.name,
 	   GTS_OBJECT_CLASS (GTS_SURFACE (object)->edge_class)->info.name,
+	   GTS_POINT_CLASS (GTS_SURFACE (object)->vertex_class)->binary ?
+	   "GtsVertexBinary" :
 	   GTS_OBJECT_CLASS (GTS_SURFACE (object)->vertex_class)->info.name);
 }
 
@@ -238,8 +241,37 @@ guint gts_surface_read (GtsSurface * surface, GtsFile * f)
     return f->line;
   }
   nf = atoi (f->token->str);
-
-  gts_file_first_token_after (f, '\n');
+  
+  gts_file_next_token (f);
+  if (f->type == GTS_STRING) {
+    if (f->type != GTS_STRING) {
+      gts_file_error (f, "expecting a string (GtsSurfaceClass)");
+      return f->line;
+    }
+    gts_file_next_token (f);
+    if (f->type != GTS_STRING) {
+      gts_file_error (f, "expecting a string (GtsFaceClass)");
+      return f->line;
+    }
+    gts_file_next_token (f);
+    if (f->type != GTS_STRING) {
+      gts_file_error (f, "expecting a string (GtsEdgeClass)");
+      return f->line;
+    }
+    gts_file_next_token (f);
+    if (f->type != GTS_STRING) {
+      gts_file_error (f, "expecting a string (GtsVertexClass)");
+      return f->line;
+    }
+    if (!strcmp (f->token->str, "GtsVertexBinary"))
+      GTS_POINT_CLASS (surface->vertex_class)->binary = TRUE;
+    else {
+      GTS_POINT_CLASS (surface->vertex_class)->binary = FALSE;
+      gts_file_first_token_after (f, '\n');
+    }
+  }
+  else
+    gts_file_first_token_after (f, '\n');
 
   if (nf <= 0)
     return 0;
@@ -255,12 +287,17 @@ guint gts_surface_read (GtsSurface * surface, GtsFile * f)
 
     (* GTS_OBJECT_CLASS (surface->vertex_class)->read) (&new_vertex, f);
     if (f->type != GTS_ERROR) {
-      gts_file_first_token_after (f, '\n');
+      if (!GTS_POINT_CLASS (surface->vertex_class)->binary)
+	gts_file_first_token_after (f, '\n');
       vertices[n++] = GTS_VERTEX (new_vertex);
     }
+    else
+      gts_object_destroy (new_vertex);
   }
   if (f->type == GTS_ERROR)
     nv = n;
+  if (GTS_POINT_CLASS (surface->vertex_class)->binary)
+    gts_file_first_token_after (f, '\n');
 
   n = 0;
   while (n < ne && f->type != GTS_ERROR) {
@@ -672,7 +709,8 @@ void gts_surface_print_stats (GtsSurface * s, FILE * fptr)
 static void write_vertex (GtsPoint * p, gpointer * data)
 {
   (*GTS_OBJECT (p)->klass->write) (GTS_OBJECT (p), (FILE *) data[0]);
-  fputc ('\n', (FILE *) data[0]);
+  if (!GTS_POINT_CLASS (GTS_OBJECT (p)->klass)->binary)
+    fputc ('\n', (FILE *) data[0]);
   g_hash_table_insert (data[2], p, 
 		       GUINT_TO_POINTER (++(*((guint *) data[1]))));
 }
@@ -753,6 +791,8 @@ void gts_surface_write (GtsSurface * s, FILE * fptr)
   n = 0;
   gts_surface_foreach_vertex (s, (GtsFunc) write_vertex, data);
   n = 0;
+  if (GTS_POINT_CLASS (s->vertex_class)->binary)
+    fputc ('\n', fptr);
   gts_surface_foreach_edge (s, (GtsFunc) write_edge, data);
   gts_surface_foreach_face (s, (GtsFunc) write_face, data);
   g_hash_table_destroy (vindex);
@@ -822,6 +862,59 @@ void gts_surface_write_oogl (GtsSurface * s, FILE * fptr)
   gts_surface_foreach_vertex (s, (GtsFunc) gts_object_reset_reserved, NULL);
 }
 
+static void write_vertex_vtk (GtsPoint * p, gpointer * data)
+{
+  FILE * fp = data[0];
+
+  fprintf (fp, "%g %g %g\n", p->x, p->y, p->z);
+  GTS_OBJECT (p)->reserved = GUINT_TO_POINTER ((*((guint *) data[1]))++);
+}
+
+static void write_face_vtk (GtsTriangle * t, FILE * fp)
+{
+  GtsVertex * v1, * v2, * v3;
+  gts_triangle_vertices (t, &v1, &v2, &v3);
+  fprintf (fp, "3 %u %u %u\n",
+	   GPOINTER_TO_UINT (GTS_OBJECT (v1)->reserved),
+	   GPOINTER_TO_UINT (GTS_OBJECT (v2)->reserved),
+	   GPOINTER_TO_UINT (GTS_OBJECT (v3)->reserved));
+}
+
+/**
+ * gts_surface_write_vtk:
+ * @s: a #GtsSurface.
+ * @fptr: a file pointer.
+ * 
+ * Writes in the file @fptr a VTK representation of @s.
+ */
+void gts_surface_write_vtk (GtsSurface * s, FILE * fptr)
+{
+  guint n = 0;
+  gpointer data[2];
+  GtsSurfaceStats stats;
+
+  g_return_if_fail (s != NULL);
+  g_return_if_fail (fptr != NULL);
+
+  data[0] = fptr;
+  data[1] = &n;
+
+  gts_surface_stats (s, &stats);
+  fprintf (fptr,
+	   "# vtk DataFile Version 2.0\n"
+	   "Generated by GTS\n"
+           "ASCII\n"
+	   "DATASET POLYDATA\n"
+	   "POINTS %u float\n",
+	   stats.edges_per_vertex.n);
+  gts_surface_foreach_vertex (s, (GtsFunc) write_vertex_vtk, data);
+  fprintf (fptr,
+	   "POLYGONS %u %u\n",
+	   stats.n_faces, stats.n_faces*4);
+  gts_surface_foreach_face (s, (GtsFunc) write_face_vtk, fptr);
+  gts_surface_foreach_vertex (s, (GtsFunc) gts_object_reset_reserved, NULL);  
+}
+
 static void write_edge_oogl_boundary (GtsSegment * s, gpointer * data)
 {
   if (!gts_edge_is_boundary (GTS_EDGE (s), data[1]))
@@ -2184,6 +2277,127 @@ gdouble gts_surface_volume (GtsSurface * s)
   return volume/6.;
 }
 
+static void center_of_mass_foreach_face (GtsTriangle * t,
+					 gpointer * data)
+{
+  GtsVertex * v1, * v2, * v3;
+  GtsPoint * p1, * p2, * p3;
+  gdouble x1, y1, z1, x2, y2, z2, nx, ny, nz;
+  gdouble * volume = data[0];
+  gdouble * cm = data[1];
+
+  gts_triangle_vertices (t, &v1, &v2, &v3);
+  p1 = GTS_POINT (v1);
+  p2 = GTS_POINT (v2);
+  p3 = GTS_POINT (v3);
+
+  x1 = p2->x - p1->x;
+  y1 = p2->y - p1->y;
+  z1 = p2->z - p1->z;
+
+  x2 = p3->x - p1->x;
+  y2 = p3->y - p1->y;
+  z2 = p3->z - p1->z;
+  
+  nx = y1*z2 - z1*y2;
+  ny = z1*x2 - x1*z2;
+  nz = x1*y2 - y1*x2;
+
+  cm[0] += nx*(p1->x*p1->x + p2->x*p2->x + p3->x*p3->x + 
+	       p1->x*p2->x + p1->x*p3->x + p2->x*p3->x);
+  cm[1] += ny*(p1->y*p1->y + p2->y*p2->y + p3->y*p3->y + 
+	       p1->y*p2->y + p1->y*p3->y + p2->y*p3->y);
+  cm[2] += nz*(p1->z*p1->z + p2->z*p2->z + p3->z*p3->z + 
+	       p1->z*p2->z + p1->z*p3->z + p2->z*p3->z);
+
+  *volume += nx*(p1->x + p2->x + p3->x);
+}
+
+
+/**
+ * gts_surface_center_of_mass:
+ * @s: a #GtsSurface.
+ * @cm: a #GtsVector.
+ *
+ * Fills @cm with the coordinates of the center of mass of @s.
+ *
+ * Returns: the signed volume of the domain bounded by the surface @s.
+ */
+gdouble gts_surface_center_of_mass (GtsSurface * s,
+				    GtsVector cm)
+{
+  gdouble volume = 0.;
+  gpointer data[2];
+
+  g_return_val_if_fail (s != NULL, 0.0);
+
+  data[0] = &volume;
+  data[1] = &(cm[0]);
+  cm[0] = cm[1] = cm[2] = 0.;
+  gts_surface_foreach_face (s, (GtsFunc) center_of_mass_foreach_face, data);
+  
+  if (volume != 0.) {
+    cm[0] /= 4.*volume;
+    cm[1] /= 4.*volume;
+    cm[2] /= 4.*volume;
+  }
+
+  return volume/6.;
+}
+
+static void center_of_area_foreach_face (GtsTriangle * t,
+					 gpointer * data)
+{
+  GtsVertex * v1, * v2, * v3;
+  GtsPoint * p1, * p2, * p3;
+  gdouble a;
+  gdouble * area = data[0];
+  gdouble * cm = data[1];
+
+  gts_triangle_vertices (t, &v1, &v2, &v3);
+  p1 = GTS_POINT (v1);
+  p2 = GTS_POINT (v2);
+  p3 = GTS_POINT (v3);
+
+  a = gts_triangle_area (t);
+  cm[0] += a*(p1->x + p2->x + p3->x);
+  cm[1] += a*(p1->y + p2->y + p3->y);
+  cm[2] += a*(p1->z + p2->z + p3->z);
+  *area += a;
+}
+
+
+/**
+ * gts_surface_center_of_area:
+ * @s: a #GtsSurface.
+ * @cm: a #GtsVector.
+ *
+ * Fills @cm with the coordinates of the center of area of @s.
+ *
+ * Returns: the area of surface @s.
+ */
+gdouble gts_surface_center_of_area (GtsSurface * s,
+				    GtsVector cm)
+{
+  gdouble area = 0.;
+  gpointer data[2];
+
+  g_return_val_if_fail (s != NULL, 0.0);
+
+  data[0] = &area;
+  data[1] = &(cm[0]);
+  cm[0] = cm[1] = cm[2] = 0.;
+  gts_surface_foreach_face (s, (GtsFunc) center_of_area_foreach_face, data);
+  
+  if (area != 0.) {
+    cm[0] /= 3.*area;
+    cm[1] /= 3.*area;
+    cm[2] /= 3.*area;
+  }
+
+  return area;
+}
+
 static void number_foreach (gpointer data, guint * n)
 {
   (*n)++;
@@ -2423,3 +2637,131 @@ void gts_surface_traverse_destroy (GtsSurfaceTraverse * t)
   g_free (t);
 }
 
+static void traverse_manifold (GtsTriangle * t, GtsSurface * s)
+{
+  if (g_slist_length (GTS_FACE (t)->surfaces) > 1)
+    return;
+
+  gts_surface_add_face (s, GTS_FACE (t));
+  if (g_slist_length (t->e1->triangles) == 2) {
+    if (t->e1->triangles->data != t)
+      traverse_manifold (t->e1->triangles->data, s);
+    else
+      traverse_manifold (t->e1->triangles->next->data, s);
+  }
+  if (g_slist_length (t->e2->triangles) == 2) {
+    if (t->e2->triangles->data != t)
+      traverse_manifold (t->e2->triangles->data, s);
+    else
+      traverse_manifold (t->e2->triangles->next->data, s);
+  }
+  if (g_slist_length (t->e3->triangles) == 2) {
+    if (t->e3->triangles->data != t)
+      traverse_manifold (t->e3->triangles->data, s);
+    else
+      traverse_manifold (t->e3->triangles->next->data, s);
+  }
+}
+
+static void non_manifold_edges (GtsEdge * e, gpointer * data)
+{
+  GtsSurface * s = data[0];
+  GSList ** non_manifold = data[1];
+
+  if (gts_edge_face_number (e, s) > 2) {
+    GSList * i = e->triangles;
+
+    while (i) {
+      if (gts_face_has_parent_surface (i->data, s) &&
+	  !g_slist_find (*non_manifold, i->data))
+	*non_manifold = g_slist_prepend (*non_manifold, i->data);
+      i = i->next;
+    }
+  }
+}
+
+static void traverse_boundary (GtsEdge * e, gpointer * data)
+{
+  GtsSurface * orig = data[0];
+  GSList ** components = data[1];
+  GtsFace * f = gts_edge_is_boundary (e, orig);
+
+  if (f != NULL && g_slist_length (f->surfaces) == 1) {
+    GtsSurface * s = 
+      gts_surface_new (GTS_SURFACE_CLASS (GTS_OBJECT (orig)->klass),
+		       orig->face_class,
+		       orig->edge_class,
+		       orig->vertex_class);
+    GSList * non_manifold = NULL, * i;
+    gpointer data[2];
+
+    *components = g_slist_prepend (*components, s);
+    data[0] = s;
+    data[1] = &non_manifold;
+    traverse_manifold (GTS_TRIANGLE (f), s);
+
+    gts_surface_foreach_edge (s, (GtsFunc) non_manifold_edges, data);
+    i = non_manifold;
+    while (i) {
+      gts_surface_remove_face (s, i->data);
+      i = i->next;
+    }
+    g_slist_free (non_manifold);
+  }
+}
+
+static void traverse_remaining (GtsFace * f, gpointer * data)
+{
+  GtsSurface * orig = data[0];
+  GSList ** components = data[1];
+
+  if (g_slist_length (f->surfaces) == 1) {
+    GtsSurface * s = 
+      gts_surface_new (GTS_SURFACE_CLASS (GTS_OBJECT (orig)->klass),
+		       orig->face_class,
+		       orig->edge_class,
+		       orig->vertex_class);
+    GSList * non_manifold = NULL, * i;
+    gpointer data[2];
+
+    *components = g_slist_prepend (*components, s);
+    data[0] = s;
+    data[1] = &non_manifold;
+    traverse_manifold (GTS_TRIANGLE (f), s);
+
+    gts_surface_foreach_edge (s, (GtsFunc) non_manifold_edges, data);
+    i = non_manifold;
+    while (i) {
+      gts_surface_remove_face (s, i->data);
+      i = i->next;
+    }
+    g_slist_free (non_manifold);
+  }
+}
+
+/**
+ * gts_surface_split:
+ * @s: a #GtsSurface.
+ *
+ * Splits a surface into connected and manifold components.
+ * 
+ * Returns: a list of new #GtsSurface.
+ */
+GSList * gts_surface_split (GtsSurface * s)
+{
+  gpointer data[2];
+  GSList * components = NULL;
+
+  g_return_val_if_fail (s != NULL, NULL);
+
+  data[0] = s;
+  data[1] = &components;
+
+  /* boundary components */
+  gts_surface_foreach_edge (s, (GtsFunc) traverse_boundary, data);
+
+  /* remaining components */
+  gts_surface_foreach_face (s, (GtsFunc) traverse_remaining, data);
+
+  return components;
+}
