@@ -20,10 +20,6 @@
 #include <math.h>
 #include "gts.h"
 
-#ifndef PI
-#define PI 3.14159265359
-#endif
-
 static void triangle_destroy (GtsObject * object)
 {
   GtsTriangle * triangle = GTS_TRIANGLE (object);
@@ -244,8 +240,8 @@ gdouble gts_triangles_angle (GtsTriangle * t1,
   pvz = nx1*ny2 - ny1*nx2;
 
   theta = atan2 (sqrt (pvx*pvx + pvy*pvy + pvz*pvz), 
-		 nx1*nx2 + ny1*ny2 + nz1*nz2) - PI;
-  return theta < - PI ? theta + 2.*PI : theta;
+		 nx1*nx2 + ny1*ny2 + nz1*nz2) - M_PI;
+  return theta < - M_PI ? theta + 2.*M_PI : theta;
 }
 
 /**
