@@ -40,18 +40,21 @@ typedef enum { NONE,
 	       FOLD, 
 	       HEIGHT, 
 	       CURVATURE, 
-	       GAUSSIAN } ColorOptions;
+	       GAUSSIAN,
+               COMPONENT } ColorOptions;
 
 ColorOptions color = NONE;
 gboolean autoscale = TRUE;
 gboolean boundary = FALSE;
 gboolean non_manifold = FALSE;
+gboolean duplicate = FALSE;
 gint flatten = -1;
 guint height = 0;
 gdouble avg = 0.;
 gdouble min = 0.0, max = 1.0;
 gdouble maxcosine2 = 1.0;
 gboolean verbose = FALSE;
+gboolean faces = FALSE;
 
 typedef struct {
   gdouble r, g, b;
@@ -238,6 +241,19 @@ static void foreach_non_manifold_edge (GtsEdge * e, gpointer info)
   }
 }
 
+static void foreach_duplicate_edge (GtsSegment * s, gpointer info)
+{
+  FILE * fp = info;
+
+  if (gts_segment_is_duplicate (s)) {
+    fputs ("VECT 1 2 0 2 0 ", fp);
+    write_point (GTS_POINT (s->v1), fp);
+    fputc (' ', fp);
+    write_point (GTS_POINT (s->v2), fp);
+    fputc ('\n', fp);
+  }
+}
+
 static void foreach_triangle (GtsTriangle * t, gpointer * info)
 {
   FILE * fp = info[0];
@@ -518,13 +534,13 @@ static GSList * chain_segments (GSList * segments)
 static void write_isolines (GtsSurface * surface, 
 			    GNode * stree,
 			    gdouble level,
-			    FILE * fp)
+			    FILE * fp,
+			    gboolean gnuplot)
 {
   GNode * ptree;
   GtsSurface * plane;
   GtsVertex * v1, * v2, * v3;
-  GtsSurfaceInter * si;
-  GSList * chains, * i;
+  GSList * inter, * chains, * i;
   guint nvertices = 0;
 
   g_return_if_fail (surface != NULL);
@@ -543,9 +559,8 @@ static void write_isolines (GtsSurface * surface,
 			       gts_edge_new (gts_edge_class (), v2, v3),
 			       gts_edge_new (gts_edge_class (), v3, v1)));
   ptree = gts_bb_tree_surface (plane);
-  si = gts_surface_inter_new (gts_surface_inter_class (),
-			      surface, plane, stree, ptree, FALSE, FALSE);
-  chains = chain_segments (si->edges);
+  inter = gts_surface_intersection (surface, plane, stree, ptree);
+  chains = chain_segments (inter);
 
   i = chains;
   while (i) {
@@ -554,21 +569,23 @@ static void write_isolines (GtsSurface * surface,
     i = i->next;
   }
 
-  fprintf (fp, "VECT %d %d 0\n", g_slist_length (chains), nvertices);
-  i = chains;
-  while (i) {
-    GSList * chain = i->data;
-
-    fprintf (fp, "%d ", g_slist_length (chain));
-    i = i->next;
+  if (!gnuplot) {
+    fprintf (fp, "VECT %d %d 0\n", g_slist_length (chains), nvertices);
+    i = chains;
+    while (i) {
+      GSList * chain = i->data;
+      
+      fprintf (fp, "%d ", g_slist_length (chain));
+      i = i->next;
+    }
+    i = chains;
+    while (i) {
+      fprintf (fp, "0 ");
+      i = i->next;
+    }
+    fprintf (fp, "\n");
   }
   i = chains;
-  while (i) {
-    fprintf (fp, "0 ");
-    i = i->next;
-  }
-  fprintf (fp, "\n");
-  i = chains;
   while (i) {
     GSList * chain = i->data;
     GSList * j = chain;
@@ -580,11 +597,13 @@ static void write_isolines (GtsSurface * surface,
       fputc ('\n', fp);
       j = j->next;
     }
+    if (gnuplot)
+      fputc ('\n', fp);
     g_slist_free (chain);
     i = i->next;
   }
   g_slist_free (chains);
-  gts_object_destroy (GTS_OBJECT (si));
+  g_slist_free (inter);
   gts_bb_tree_destroy (ptree, TRUE);
   gts_object_destroy (GTS_OBJECT (plane));
 }
@@ -594,6 +613,31 @@ static void sum_flatten (GtsPoint * p)
   avg += (&p->x)[flatten];
 }
 
+static void write_face (GtsTriangle * t)
+{
+  GtsVertex * v1, * v2, * v3;
+  GtsPoint o;
+ 
+  gts_triangle_vertices (t, &v1, &v2, &v3);
+  o.x = (GTS_POINT (v1)->x + GTS_POINT (v2)->x + GTS_POINT (v3)->x)/3.;
+  o.y = (GTS_POINT (v1)->y + GTS_POINT (v2)->y + GTS_POINT (v3)->y)/3.;
+  o.z = (GTS_POINT (v1)->z + GTS_POINT (v2)->z + GTS_POINT (v3)->z)/3.;
+  printf ("OFF 3 1 3\n"
+	  "%g %g %g\n"
+	  "%g %g %g\n"
+	  "%g %g %g\n"
+	  "3 0 1 2\n",
+	  o.x + 0.9*(GTS_POINT (v1)->x - o.x),
+	  o.y + 0.9*(GTS_POINT (v1)->y - o.y),
+	  o.z + 0.9*(GTS_POINT (v1)->z - o.z),
+	  o.x + 0.9*(GTS_POINT (v2)->x - o.x),
+	  o.y + 0.9*(GTS_POINT (v2)->y - o.y),
+	  o.z + 0.9*(GTS_POINT (v2)->z - o.z),
+	  o.x + 0.9*(GTS_POINT (v3)->x - o.x),
+	  o.y + 0.9*(GTS_POINT (v3)->y - o.y),
+	  o.z + 0.9*(GTS_POINT (v3)->z - o.z));
+}
+
 int main (int argc, char * argv[])
 {
   GtsSurface * surface;
@@ -602,15 +646,22 @@ int main (int argc, char * argv[])
   guint isolines = 0;
   GtsFile * fp;
   gdouble feature = 0.;
+  gdouble isoline[256];
+  guint niso = 0;
+  gboolean gnuplot = FALSE;
 
   colormap = colormap_jet (); /* default */
   while (c != EOF) {
 #ifdef HAVE_GETOPT_LONG
     static struct option long_options[] = {
+      {"gnuplot", no_argument, NULL, 'G'},
+      {"faces", no_argument, NULL, 't'},
+      {"isoline", required_argument, NULL, 'I'},
       {"gaussian", no_argument, NULL, 'g'},
       {"curvature", no_argument, NULL, 'C'},
       {"feature", required_argument, NULL, 'e'},
       {"flatten", required_argument, NULL, 'F'},
+      {"component", no_argument, (int *)&color, COMPONENT},
       {"quality", no_argument, (int *)&color, QUALITY},
       {"area", no_argument, (int *)&color, AREA},
       {"incomp", no_argument, (int *)&color, INCOMP},
@@ -619,6 +670,7 @@ int main (int argc, char * argv[])
       {"height", required_argument, NULL, 'H'},
       {"boundary", no_argument, &boundary, TRUE},
       {"non-manifold", no_argument, &non_manifold, TRUE},
+      {"duplicate", no_argument, &duplicate, TRUE},
       {"isolines", required_argument, NULL, 'i'},
       {"cmap", required_argument, NULL, 'c'},
       {"help", no_argument, NULL, 'h'},
@@ -629,11 +681,20 @@ int main (int argc, char * argv[])
       {"verbose", no_argument, NULL, 'v'}
     };
     int option_index = 0;
-    switch ((c = getopt_long (argc, argv, "c:hm:M:rvf:i:nF:e:H:Cg",
+    switch ((c = getopt_long (argc, argv, "c:hm:M:rvf:i:nF:e:H:CgI:tG",
 			      long_options, &option_index))) {
 #else /* not HAVE_GETOPT_LONG */
-    switch ((c = getopt (argc, argv, "c:hm:M:rvf:i:nF:e:H:Cg"))) {
+    switch ((c = getopt (argc, argv, "c:hm:M:rvf:i:nF:e:H:CgI:tG"))) {
 #endif /* not HAVE_GETOPT_LONG */
+    case 'G': /* gnuplot */
+      gnuplot = TRUE;
+      break;
+    case 't': /* faces */
+      faces = TRUE;
+      break;
+    case 'I': /* isoline */
+      isoline[niso++] = atof (optarg);
+      break;
     case 'g': /* gaussian */
       color = GAUSSIAN;
       break;
@@ -693,11 +754,15 @@ int main (int argc, char * argv[])
 	     "Convert a GTS file (surface file format generated by the Gts Library)\n"
 	     "to OOGL file format (Geomview).\n"
 	     "\n"
+             "  -G      --gnuplot      writes isolines in gnuplot format\n"   
+             "          --component    color faces according to the component\n"
+             "                         they belong too\n"
 	     "          --quality      color faces according to their quality\n"
 	     "          --area         color faces according to their area\n"
 	     "          --incomp       color incompatible faces\n"
 	     "  -f VAL  --fold=VAL     color faces which make an angle smaller\n" 
 	     "                         than VAL degrees with any of their neighbors\n"
+             "  -t      --faces        output individual faces\n"
 	     "          --epv          color vertices according to number of edges per vertex\n"
 	     "  -H C    --height=C     color vertices according to their C coordinate\n"
 	     "  -g      --gaussian     color vertices according to Gaussian curvature\n"
@@ -705,7 +770,9 @@ int main (int argc, char * argv[])
 	     "          --boundary     output boundary edges\n"
              "  -e A    --feature=A    output `feature' edges defined by angle A\n"
              "          --non-manifold output non-manifold edges\n"
+             "          --duplicate    output duplicate edges\n"
 	     "  -i N    --isolines=N   draw N isolines (levels of constant altitude)\n"
+             "  -I L    --isolines=L   draw isoline at level L\n"
 	     "          --cmap=FILE    load FILE as colormap\n"
 	     "  -m VAL, --min=VAL      use VAL as minimum scaling value\n"
 	     "  -M VAL, --max=VAL      use VAL as maximum scaling value\n"
@@ -749,40 +816,51 @@ int main (int argc, char * argv[])
       avg /= gts_surface_vertex_number (surface);
   }
 
-  if (boundary || non_manifold || isolines || feature > 0.) {
+  if (boundary || non_manifold || duplicate || 
+      isolines || feature > 0. || niso > 0) {
     if (!nosurface) {
       puts ("(geometry \"surface\" { = ");
       oogl_surface (surface, stdout);
       puts ("})");
     }
     if (boundary) {
-      if (!nosurface || isolines || non_manifold)
+      if (!nosurface || isolines || niso > 0 || non_manifold || duplicate)
 	puts ("(geometry \"boundary\" { = ");
       puts ("LIST {");
       gts_surface_foreach_edge (surface, (GtsFunc) foreach_boundary_edge, 
 				stdout);
       puts ("}");
-      if (!nosurface || isolines || non_manifold)
+      if (!nosurface || isolines || niso > 0 || non_manifold || duplicate)
 	puts ("})");
     }
     if (feature > 0.) {
-      if (!nosurface || isolines || non_manifold)
+      if (!nosurface || isolines || niso > 0 || non_manifold || duplicate)
 	puts ("(geometry \"feature\" { = ");
       puts ("LIST {");
       gts_surface_foreach_edge (surface, (GtsFunc) foreach_feature_edge, 
 				&feature);
       puts ("}");
-      if (!nosurface || isolines || non_manifold)
+      if (!nosurface || isolines || niso > 0 || non_manifold || duplicate)
 	puts ("})");
     }
     if (non_manifold) {
-      if (!nosurface || isolines || boundary)
+      if (!nosurface || isolines || niso > 0 || boundary || duplicate)
 	puts ("(geometry \"non-manifold\" { = ");
       puts ("LIST {");
       gts_surface_foreach_edge (surface, (GtsFunc) foreach_non_manifold_edge, 
 				stdout);
       puts ("}");
-      if (!nosurface || isolines || boundary)
+      if (!nosurface || isolines || niso > 0 || boundary || duplicate)
+	puts ("})");
+    }
+    if (duplicate) {
+      if (!nosurface || isolines || niso > 0 || boundary || non_manifold)
+	puts ("(geometry \"duplicate\" { = ");
+      puts ("LIST {");
+      gts_surface_foreach_edge (surface, (GtsFunc) foreach_duplicate_edge, 
+				stdout);
+      puts ("}");
+      if (!nosurface || isolines || niso > 0 || boundary || non_manifold)
 	puts ("})");
     }
     if (isolines > 0) {
@@ -797,19 +875,70 @@ int main (int argc, char * argv[])
       z = bbox->z1 + dz;
       if (!nosurface || boundary)
 	puts ("(geometry \"isolines\" { = ");
-      puts ("LIST {");
+      if (!gnuplot)
+	puts ("LIST {");
       for (i = 0; i < isolines; i++, z += dz)
-	write_isolines (surface, stree, z, stdout);
-      puts ("}");
+	write_isolines (surface, stree, z, stdout, gnuplot);
+      if (!gnuplot)
+	puts ("}");
+      if (!nosurface || boundary)
+	puts ("})\n");
+      gts_bb_tree_destroy (stree, TRUE);
+    }
+    else if (niso > 0) {
+      GNode * stree;
+      GtsBBox * bbox;
+      guint i;
+
+      stree = gts_bb_tree_surface (surface);
+      bbox = stree->data;
+      if (!nosurface || boundary)
+	puts ("(geometry \"isolines\" { = ");
+      if (!gnuplot)
+	puts ("LIST {");
+      for (i = 0; i < niso; i++)
+	write_isolines (surface, stree, isoline[i], stdout, gnuplot);
+      if (!gnuplot)
+	puts ("}");
       if (!nosurface || boundary)
 	puts ("})\n");
       gts_bb_tree_destroy (stree, TRUE);
     }
   }
+  else if (color == COMPONENT) {
+    GSList * i = gts_surface_split (surface);
+
+    if (verbose)
+      fprintf (stderr, "%u components\n", g_slist_length (i));
+
+    printf ("LIST {\n");
+    while (i) {
+      gdouble 
+	r = rand()/(gdouble)RAND_MAX, 
+	g = rand()/(gdouble)RAND_MAX, 
+	b = rand()/(gdouble)RAND_MAX;
+
+      printf ("geom { appearance { material {\n"
+	      "  ambient %g %g %g\n"
+	      "  diffuse %g %g %g\n"
+	      "}}\n", 
+	      r, g, b, r, g, b);
+      gts_surface_write_oogl (i->data, stdout);
+      printf ("}\n");
+      i = i->next;
+    }
+    printf ("}\n");
+  }
   else if (!nosurface) {
     if (min == max) 
       max = min + 1.;
-    oogl_surface (surface, stdout);
+    if (faces) {
+      printf ("LIST {\n");
+      gts_surface_foreach_face (surface, (GtsFunc) write_face, NULL);
+      printf ("}\n");
+    }      
+    else	
+      oogl_surface (surface, stdout);
   }
 
   return 0;
